from typing import Union, TYPE_CHECKING, List, Dict, Callable
from warnings import warn

# TODO: this module depends on pandas dataframes. Should it be set as package requirement?
from pandas import DataFrame, concat

from .milp_bool import milpBool
from .sim_bool import SimBool
from .analysis_utils import decode_solver_solution
from mewpy.mew.algebra import Symbolic
from mewpy.mew.variables import Regulator

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


# TODO: type hinting and documentation
def slim_milp_bool(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                   milp_bool: milpBool = None,
                   objective: Union[str, Dict[str, Union[float, int]]] = None,
                   minimize: bool = False,
                   initial_state: Dict[str, Union[float, int]] = None,
                   get_values: bool = False) -> Union[int, float, None]:

    if not milp_bool:
        milp_bool = milpBool(model, build=True, attach=False)

    # noinspection PyProtectedMember
    old_linear_obj = milp_bool._linear_objective.copy()
    # noinspection PyProtectedMember
    old_quadratic_obj = milp_bool._quadratic_objective.copy()
    old_sense = milp_bool.minimize

    sol = milp_bool.optimize(initial_state=initial_state,
                             objective=objective,
                             minimize=minimize,
                             to_solver=True,
                             get_values=get_values)

    milp_bool.set_objective(linear=old_linear_obj, quadratic=old_quadratic_obj, minimize=old_sense)

    return decode_solver_solution(solution=sol, minimize=minimize, status=False), sol.values


def slim_sim_bool(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                  sim_bool: SimBool = None,
                  initial_state: Dict[str, Union[float, int]] = None) -> Union[int,
                                                                               float,
                                                                               None]:
    if not sim_bool:
        sim_bool = SimBool(model, build=True, attach=False)

    return sim_bool.optimize(initial_state=initial_state, to_dict=True)


def single_regulator_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                              regulators: List[Union[str, Regulator]] = None,
                              initial_state: Dict[str, Union[float, int]] = None,
                              to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Single regulator deletions

    :param model:
    :param regulators:
    :param initial_state:
    :param to_dict:

    :return:
    """

    if not initial_state:
        initial_state = {}

    reg_state = {reg.id: reg.coefficient.active_coefficient
                 for reg in model.yield_regulators()}

    reg_state.update(initial_state)

    res = {}

    if not regulators:
        regulators = model.yield_regulators()

    for regulator in regulators:

        if isinstance(regulator, str):

            regulator_obj = model.get(regulator)

        else:
            regulator_obj = regulator

        reg_state[regulator_obj.id] = 0.0

        dfs = []

        for interaction in regulator.yield_interactions():

            for coefficient, regulatory_event in interaction.regulatory_events.items():

                if not regulatory_event.is_none:

                    values = {key: val for key, val in reg_state.items()
                              if key in regulatory_event.variables}

                    df = regulatory_event.truth_table(values=values,
                                                      active_states=True,
                                                      coefficient=coefficient,
                                                      default=0.0)

                    df.index = [interaction.target.id] * df.shape[0]

                    dfs.append(df)

        df = concat(dfs)

        df = df[['result'] + [col for col in df.columns if col != 'result']]

        del reg_state[regulator_obj.id]

        res[regulator_obj.id] = df

    if to_dict:
        return res

    return concat(res.values(), keys=res.keys())


def regulatory_events(model,
                      initial_state: Dict[str, Union[float, int]] = None,
                      active_states: bool = True,
                      operators: Dict[Symbolic, Callable] = None,
                      decoder: dict = None):
    """

    Regulatory events of a model

    :param model:
    :param initial_state:
    :param active_states:
    :param operators:
    :param decoder:
    :return:
    """

    if not active_states:
        warn('Attention! Active states was set to False without an initial state. '
             'Note that computing all possible states may take some time for large models')

    dfs = []

    for interaction in model.yield_interactions():

        for coefficient, regulatory_event in interaction.regulatory_events.items():

            if not regulatory_event.is_none:
                df = regulatory_event.truth_table(values=initial_state,
                                                  active_states=active_states,
                                                  coefficient=coefficient,
                                                  operators=operators,
                                                  default=0.0,
                                                  decoder=decoder)

                df.index = [interaction.target.id] * df.shape[0]

                dfs.append(df)

    df = concat(dfs)

    df = df[['result'] + [col for col in df.columns if col != 'result']]

    return df



import warnings
from collections import OrderedDict
import numpy as np
from reframed.core.model import Model
from mewpy.util.parsing import Arithmetic, build_tree
import copy


def calculate_yprime(y, rate, substrates, products, substrate_names):
    """
    This function is used by the rate classes the user creates.
    It takes the numpy array for y_prime,
    and adds or subtracts the amount in rate to all the substrates or products listed
    Returns the new y_prime
    Args:
        y: a numpy array for the substrate values, the same order as y
        rate: the rate calculated by the user made rate equation
        substrates: list of substrates for which rate should be subtracted
        products: list of products for which rate should be added
        substrate_names: the ordered list of substrate names in the model.\
            Used to get the position of each substrate or product in y_prime
    Returns:
        y_prime: following the addition or subtraction of rate to the specificed substrates
    """
    y_prime = np.zeros(len(y))

    for name in substrates:
        y_prime[substrate_names.index(name)] -= rate

    for name in products:
        y_prime[substrate_names.index(name)] += rate

    return y_prime


def check_positive(y_prime):
    """
    Check that substrate values are not negative when they shouldnt be.
    """

    for i in range(len(y_prime)):
        if y_prime[i] < 0:
            y_prime[i] = 0

    return y_prime


class Rule(object):
    """Base class for kinetic rules.
    """

    def __init__(self, r_id, formula: str, parameters: dict = {}):
        """Creates a new rule

        Args:
            r_id (str): Reaction/rule identifier
            formula (str): The rule string representation.
        """
        self.id = r_id
        self.formula = formula
        self._tree = None
        self.parameters = parameters

    @property
    def tree(self):
        """Parsing tree of the formula.

        Returns:
            Node: Root node of the parsing tree.
        """
        if not self._tree:
            self._tree = build_tree(self.formula, Arithmetic)
        return self._tree

    def parse_parameters(self):
        """Returns the list of parameters within the rule.

        Returns:
            list: parameters
        """
        return list(self.tree.get_parameters())

    def get_parameters(self):
        return self.parameters

    def replace(self, parameters=None, local=True, infix=True):
        """Replaces parameters with values taken from a dictionary.
        If no parameter are given for replacement, returns the string representation of the rule
        built from the parsing tree.

        Args:
            parameters (dict, optional): Replacement dictionary. Defaults to None.

        Returns:
            str: the kinetic rule.
        """
        param = parameters.copy()
        if local:
            param.update(self.parameters)
        t = self.tree.replace(param)
        if infix:
            return t.to_infix()
        else:
            return t

    def calculate_rate(self, substrates={}, parameters={}):
        param = {}
        param.update(substrates)
        param.update(parameters)
        param.update(self.parameters)
        if len(param.keys()) != len(self.parse_parameters()):
            s = set(self.parse_parameters())-set(param.keys())
            raise ValueError(f"Values missing for parameters: {s}")
        t = self.replace(param)
        rate = eval(t)
        return rate

    def __str__(self):
        return self.formula

    def __repr__(self):
        return self.replace()


class KineticReaction(Rule):

    def __init__(self, r_id, formula: str, substrates: list = [],
                 products: list = [], parameters: dict = {}, modifiers: list = []):
        """[summary]

        Args:
            r_id (str): Reaction identifier
            formula (str): kinetic law
            parameters (dict, optional): local parameters. Defaults to dict().
            substrates (list, optional): substrates. Defaults to [].
            products (list, optional): products. Defaults to [].
        """
        super(KineticReaction, self).__init__(r_id, formula, parameters)
        self.substrates = substrates
        self.products = products
        self.modifiers = modifiers
        self.parameter_distributions = {}

    def parse_law(self, map: dict, local=True):
        """Auxialiary method invoked by the model to build the ODE system.

        Args:
            map (dict): Dictionary of global paramameters replacements.

        Returns:
            str: kinetic rule.
        """
        m = {p_id: f"p['{self.id}_{p_id}']" for p_id in self.parameters.keys()}
        r_map = map.copy()
        r_map.update(m)
        return self.replace(r_map,local=local)

    def calculate_rate(self, substrates={}, parameters={}):
        param = {}
        param.update(substrates)
        param.update(parameters)
        param.update(self.parameters)
        if len(param.keys()) != len(self.parse_parameters()):
            s = set(self.parse_parameters())-set(param.keys())
            r = s - set(self.parameter_distributions.keys())
            if r:
                raise ValueError(f"Missing values or distribuitions for parameters: {r}")
            else:
                for p in s:
                    param[p] = self.parameter_distributions[p].rvs()
        t = self.replace(param)
        rate = eval(t)
        return rate

    def reaction(self, y, substrate_names, parameter_dict):

        if self.substrate_indexes == []:
            # need to move this to the model
            self.get_indexes(substrate_names)

        if self.run_model_parameters == []:
            self.run_model_parameters = self.get_parameters(parameter_dict)

        for modifier in self.modifiers:
            if modifier.substrate_indexes == []:
                modifier.get_substrate_indexes(self.reaction_substrate_names)
            if modifier.parameter_indexes == []:
                modifier.get_parameter_indexes(self.parameter_names)

        substrates = self.get_substrates(y)
        parameters = copy.copy(self.run_model_parameters)

        if len(self.modifiers) != 0:
            substrates, parameters = self.calculate_modifiers(substrates, parameters)

        rate = self.calculate_rate(substrates, parameters)

        y_prime = calculate_yprime(y, rate, self.substrates, self.products, substrate_names)
        y_prime = self.modify_product(y_prime, substrate_names)

        if self.check_positive:
            y_prime = check_positive(y_prime)

        return y_prime

    def set_parameter_defaults_to_mean(self):
        """Sets not defined parameters to the median of a distribution.
        """
        for name in self.parameter_distributions:
            if name not in self.parameters:
                if (type(self.parameter_distributions[name]) == list or
                        type(self.parameter_distributions[name]) == tuple):
                    self.parameters[name] = (self.parameter_distributions[name][0] +
                                             self.parameter_distributions[name][1]) / 2
                else:
                    self.parameters[name] = self.parameter_distributions[name].mean()



class ODEModel(Model):
    # TODO: Generalize to work with any model
    def __init__(self, model_id):
        """ ODE Model.
        """
        super(ODEModel, self).__init__(model_id)
        # kinetic rule of each reaction
        self.ratelaws = OrderedDict()
        # initial concentration of metabolites
        self.concentrations = OrderedDict()
        # parameter defined as constantes
        self.constant_params = OrderedDict()
        # variable parameters
        self.variable_params = OrderedDict()
        self.assignment_rules = OrderedDict()

        self._func_str = None
        self._constants = None

        self.reacParamsFactors = None
        self.parsedRates = None
        self.parsedRules = None
        self.parsedXdot = None

    def get_reactions(self):
        return self.reactions

    def _clear_temp(self):
        self.update()
        self._func_str = None

    def add_reaction(self, reaction, replace=True):
        super(ODEModel, self).add_reaction(reaction, replace)

    def set_concentration(self, m_id: str, concentration: float):
        """Sets a metabolite initial concentration

        Args:
            m_id (str): Metabolite identifier
            concentration (float): Initial concentration
        """
        if m_id in self.metabolites:
            self.concentrations[m_id] = concentration
        else:
            warnings.warn(f"No such metabolite '{m_id}'", RuntimeWarning)

    def set_ratelaw(self, r_id: str, law: KineticReaction):
        """Define the rate law for a given reaction.

        Args:
            r_id (str): Reaction Identifier
            law (KineticReaction): The reaction rate law.
        """
        if r_id in self.reactions:
            self.ratelaws[r_id] = law
        else:
            warnings.warn(f"No such reaction '{r_id}'", RuntimeWarning)

    def set_assignment_rule(self, p_id: str, rule: Rule):
        if p_id in self.variable_params or p_id in self.metabolites:
            self.assignment_rules[p_id] = rule
        else:
            warnings.warn(f"No such variable parameter '{p_id}'", RuntimeWarning)

    def set_global_parameter(self, key, value, constant=True):
        if constant:
            self.constant_params[key] = value
        else:
            self.variable_params[key] = value

    def remove_reactions(self, id_list):
        super(ODEModel, self).remove_reactions(id_list)
        for r_id in id_list:
            del self.ratelaws[r_id]

    def merge_constants(self):
        constants = OrderedDict()

        for c_id, comp in self.compartments.items():
            constants[c_id] = comp.size

        constants.update(self.constant_params)

        for r_id, law in self.ratelaws.items():
            for p_id, value in law.parameters.items():
                full_id = f"{r_id}_{p_id}"
                constants[full_id] = value

        self._constants = constants
        return constants

    def print_balance(self, m_id):
        c_id = self.metabolites[m_id].compartment
        table = self.metabolite_reaction_lookup()
        terms = [f"{coeff:+g} * r['{r_id}']" for r_id, coeff in table[m_id].items()]
        if len(terms) == 0 or (self.metabolites[m_id].constant and self.metabolites[m_id].boundary):
            expr = "0"
        else:
            expr = f"1/p['{c_id}'] * ({' '.join(terms)})"
        return expr

    def get_parameters(self, exclude_compartments=False):
        if not self._constants:
            self.merge_constants()
        parameters = self._constants.copy()
        if exclude_compartments:
            for c_id in self.compartments:
                del parameters[c_id]
        return parameters

    def deriv(self, y, t):
        m_y = OrderedDict(zip(self.metabolites, y))
        yprime = np.zeros(len(y))
        for _, reaction in self.ratelaws.items():
            yprime += reaction.reaction(m_y, self.get_parameters())
        return yprime

    def build_ode(self,factors=None, local=False):
        rmap = OrderedDict()
        m = {m_id: f"x[{i}]" for i, m_id in enumerate(self.metabolites)}
        c = {c_id: f"p['{c_id}']" for c_id in self.compartments}
        p = {p_id: f"p['{p_id}']" for p_id in self.constant_params}
        v = {p_id: f"v['{p_id}']" for p_id in self.variable_params}
        rmap.update(m)
        rmap.update(c)
        rmap.update(p)
        rmap.update(v)

        parsed_rates = {r_id: ratelaw.parse_law(rmap,local=local)
                        for r_id, ratelaw in self.ratelaws.items()}        
        r = {r_id: f"({parsed_rates[r_id]})" for r_id in self.reactions}
        rmap.update(r)
       
        rate_exprs = [' '*4+"r['{}'] = {}".format(r_id, parsed_rates[r_id])
                      for r_id in self.reactions]

        balances = [' '*8 + self.print_balance(m_id) for m_id in self.metabolites]

        func_str = 'def ode_func(t, x, r, p, v):\n\n' + \
            '\n'.join(rate_exprs) + '\n\n' + \
            '    dxdt = [\n' + \
            ',\n'.join(balances) + '\n' + \
            '    ]\n\n' + \
            '    return dxdt\n'

        self._func_str = func_str

        return self._func_str

    def get_ode(self, r_dict=None, params=None, factors=None):

        p = self.merge_constants()
        v = self.variable_params.copy()

        if r_dict is not None:
            r = r_dict
        else:
            r = {}

        if params:
            p.update(params)

        exec('from math import log', globals())
        exec(self.build_ode(factors), globals())
        ode_func = eval('ode_func')

        return lambda t, x: ode_func(t, x, r, p, v)



from collections import defaultdict
from typing import Union, TYPE_CHECKING, List, Dict, Tuple

# TODO: this module depends on pandas dataframes. Should it be set as package requirement?
# noinspection PyPackageRequirements
from pandas import DataFrame

from mewpy.util.constants import ModelConstants
from mewpy.mew.variables import Reaction, Gene
from .fba import FBA, pFBA, milpFBA
from .analysis_utils import decode_solver_solution

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


# TODO: missing documentation and typing
def _fba(method, objective, minimize, constraints, status=False):

    # noinspection PyProtectedMember
    old_linear_obj = method._linear_objective.copy()
    # noinspection PyProtectedMember
    old_quadratic_obj = method._quadratic_objective.copy()
    old_sense = method.minimize

    sol = method.optimize(objective=objective,
                          minimize=minimize,
                          constraints=constraints,
                          to_solver=True,
                          get_values=False)

    method.set_objective(linear=old_linear_obj, quadratic=old_quadratic_obj, minimize=old_sense)

    return decode_solver_solution(solution=sol, minimize=minimize, status=status)


def slim_fba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
             fba: FBA = None,
             objective: Union[str, Dict[str, Union[float, int]]] = None,
             minimize: bool = False,
             constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None) -> Union[int, float, None]:

    if not fba:
        fba = FBA(model, build=True, attach=False)

    return _fba(method=fba,
                objective=objective,
                minimize=minimize,
                constraints=constraints)


def slim_pfba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
              pfba: pFBA = None,
              objective: Union[str, Dict[str, Union[float, int]]] = None,
              minimize: bool = False,
              constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None) -> Union[int, float, None]:
    if not pfba:
        pfba = pFBA(model, build=True, attach=False)

    return _fba(method=pfba,
                objective=objective,
                minimize=minimize,
                constraints=constraints)


def slim_milp_fba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                  milp_fba: milpFBA = None,
                  objective: Union[str, Dict[str, Union[float, int]]] = None,
                  minimize: bool = False,
                  constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None) -> Union[int,
                                                                                                       float,
                                                                                                       None]:
    if not milp_fba:
        milp_fba = pFBA(model, build=True, attach=False)

    return _fba(method=milp_fba,
                objective=objective,
                minimize=minimize,
                constraints=constraints)


def _inputs_processing(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                       method: str = None,
                       fraction: float = None,
                       reactions: List[Union[str, Reaction]] = None,
                       genes: List[Union[str, Gene]] = None,
                       objective: Union[str, Dict[str, Union[float, int]]] = None,
                       constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None):
    methods_map = {'fba': FBA,
                   'milp': milpFBA,
                   'pfba': pFBA}

    if not constraints:
        constraints = {}

    if objective:

        if isinstance(objective, str):

            rxn = model.get(objective)

            if not rxn:
                ValueError(f'The {objective} objective is not available in the {model} model')

            objective = {objective: 1.0}

        elif isinstance(objective, dict):

            initial_obj = objective

            objective = {}

            for key in initial_obj:

                if isinstance(key, tuple):
                    ValueError('Quadratic objectives are not allowed')

                rxn = model.get(key)

                if not rxn:
                    ValueError(f'The {key} objective is not available in the {model} model')

        else:
            raise TypeError(f'Objective wrong type {type(objective)}')

    else:
        objective = {rxn.id: coef for rxn, coef in model.objective.items()}

        if len(objective) > 1:
            raise ValueError('linear objectives having more than one variable or quadratic objectives are not allowed')

    obj_rxn = next(iter(objective.keys()))

    _FBA = methods_map.get(method, None)

    if _FBA is None:
        raise ValueError(f'Simulation method {method} is not valid. Available methods: {methods_map}')

    fba = _FBA(model, build=True, attach=False)
    fba.set_objective(linear=objective, minimize=False)

    if fraction is not None:

        if fraction > 0.0:
            sol = _fba(method=fba,
                       objective=objective,
                       minimize=False,
                       constraints=constraints)

            constraints[obj_rxn] = (fraction * sol, ModelConstants.REACTION_UPPER_BOUND)

    if not reactions:
        reactions = model.reactions.keys()

    if not genes:
        genes = model.genes.keys()

    return fba, reactions, genes, constraints


def fva(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
        method: str = 'fba',
        fraction: float = 1.0,
        reactions: List[Union[str, Reaction]] = None,
        objective: Union[str, Dict[str, Union[float, int]]] = None,
        constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None,
        to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Flux Variability Analysis (FVA)

    :param model:
    :param method:
    :param fraction:
    :param reactions:
    :param objective:
    :param constraints:
    :param to_dict:

    :return:
    """

    fba, reactions, _, constraints = _inputs_processing(model=model,
                                                        method=method,
                                                        fraction=fraction,
                                                        reactions=reactions,
                                                        genes=None,
                                                        objective=objective,
                                                        constraints=constraints)

    res = defaultdict(list)

    for rxn in reactions:

        if isinstance(rxn, str):

            rxn_objective = rxn

        else:
            rxn_objective = rxn.id

        sol = _fba(method=fba,
                   objective=rxn_objective,
                   minimize=True,
                   constraints=constraints)

        res[rxn_objective].append(sol)

        sol = _fba(method=fba,
                   objective=rxn_objective,
                   minimize=False,
                   constraints=constraints)

        res[rxn_objective].append(sol)

    if to_dict:
        return res

    return DataFrame.from_dict(data=res, orient='index', columns=['minimum', 'maximum'])


def _milp_gene_deletion(genes, constraints, fba):
    res = {}

    for gene in genes:

        if isinstance(gene, str):

            gene_id = gene

        else:
            gene_id = gene.id

        constraints[gene_id] = (0.0, 0.0)

        sol, status = _fba(method=fba,
                           objective=None,
                           minimize=False,
                           constraints=constraints,
                           status=True)

        res[gene_id] = [sol, status]

        del constraints[gene_id]

    return res


def _fba_gene_deletion(model, genes, constraints, fba):
    values = {gene.id: 1.0 for gene in model.yield_genes()}

    res = {}

    for gene in genes:

        if isinstance(gene, str):

            gene_obj = model.get(gene)

        else:
            gene_obj = gene

        values[gene_obj.id] = 0.0

        to_remove = []
        for rxn in gene_obj.yield_reactions():

            if not rxn.gpr.is_none:

                gpr_eval = rxn.gpr.evaluate(values=values)

                if not gpr_eval:
                    constraints[rxn.id] = (0.0, 0.0)
                    to_remove.append(rxn.id)

        sol, status = _fba(method=fba,
                           objective=None,
                           minimize=False,
                           constraints=constraints,
                           status=True)

        res[gene_obj.id] = [sol, status]

        for rxn in to_remove:
            del constraints[rxn]

    return res


def single_gene_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                         method: str = 'fba',
                         genes: List[Union[str, Gene]] = None,
                         objective: Union[str, Dict[str, Union[float, int]]] = None,
                         constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None,
                         to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Single gene deletions

    :param model:
    :param method:
    :param genes:
    :param objective:
    :param constraints:
    :param to_dict:

    :return:
    """
    fba, _, genes, constraints = _inputs_processing(model=model,
                                                    method=method,
                                                    fraction=None,
                                                    reactions=None,
                                                    genes=genes,
                                                    objective=objective,
                                                    constraints=constraints)

    if method == 'milp':

        res = _milp_gene_deletion(genes=genes, constraints=constraints, fba=fba)

    else:

        res = _fba_gene_deletion(model=model, genes=genes, constraints=constraints, fba=fba)

    if to_dict:
        return res

    return DataFrame.from_dict(data=res, orient='index', columns=['growth', 'status'])


def single_reaction_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                             method: str = 'fba',
                             reactions: List[Union[str, Reaction]] = None,
                             objective: Union[str, Dict[str, Union[float, int]]] = None,
                             constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None,
                             to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Single reaction deletions

    :param model:
    :param method:
    :param reactions:
    :param objective:
    :param constraints:
    :param to_dict:

    :return:
    """
    fba, reactions, _, constraints = _inputs_processing(model=model,
                                                        method=method,
                                                        fraction=None,
                                                        reactions=reactions,
                                                        genes=None,
                                                        objective=objective,
                                                        constraints=constraints)

    res = {}

    for rxn in reactions:

        if isinstance(rxn, str):

            rxn_id = rxn

        else:
            rxn_id = rxn.id

        constraints[rxn_id] = (0.0, 0.0)

        sol, status = _fba(method=fba,
                           objective=None,
                           minimize=False,
                           constraints=constraints,
                           status=True)

        res[rxn_id] = [sol, status]

        del constraints[rxn_id]

    if to_dict:
        return res

    return DataFrame.from_dict(data=res, orient='index', columns=['growth', 'status'])



from re import findall
from typing import Any, Dict, Generator, Union, TYPE_CHECKING

from mewpy.util.utilities import generator, chemical_formula_re
from mewpy.util.serialization import serialize
from mewpy.util.history import recorder
from mewpy.util.constants import atomic_weights
from .variable import Variable

if TYPE_CHECKING:
    from .reaction import Reaction


# TODO: methods stubs
class Metabolite(Variable, variable_type='metabolite', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 charge: int = None,
                 compartment: str = None,
                 formula: str = None,
                 reactions: Dict[str, 'Reaction'] = None,
                 **kwargs):

        """
        A metabolite is regularly associated with reactions and
        can usually be available as regulator too.
        It holds information regarding the charge, compartment, formula and reactions to which is associated
        Some dynamic information is inferred from the formula such as molecular weight and atoms.
        Other information is inferred from the reactions associated, such as the exchange reaction

        :param identifier: identifier, e.g. h2o_e
        :param charge: the charge of the metabolite
        :param compartment: the compartment of this respective metabolite
        :param formula: a string-like representation of the chemical formula
        :param reactions: the dictionary of reactions to which the metabolite is associated with
        """

        if not charge and charge is not 0:
            charge = None

        if not compartment:
            compartment = None

        if not formula:
            formula = ''

        if not reactions:
            reactions = {}

        self._charge = charge
        self._compartment = compartment
        self._formula = formula
        self._reactions = reactions

        super().__init__(identifier,
                         **kwargs)

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Metabolite.variable_type}

        _types.update(super(Metabolite, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):

        return f'{self.id}: {self.name} {self.formula}'

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('charge', 'charge', '_charge')
    @property
    def charge(self) -> int:

        if self._charge is None:
            return 0

        return self._charge

    @serialize('compartment', 'compartment', '_compartment')
    @property
    def compartment(self) -> str:
        return self._compartment

    @serialize('formula', 'formula', '_formula')
    @property
    def formula(self) -> str:
        return self._formula

    @serialize('reactions', 'reactions', '_reactions')
    @property
    def reactions(self) -> Dict[str, 'Reaction']:
        return self._reactions.copy()

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @charge.setter
    @recorder
    def charge(self, value):

        if not value and value is not 0:
            value = None

        self._charge = value

    @formula.setter
    @recorder
    def formula(self, value):

        if not value:
            value = ''

        self._formula = value

    @compartment.setter
    @recorder
    def compartment(self, value):

        if not value:
            value = None

        self._compartment = value

    @reactions.setter
    def reactions(self, value):

        if not value:
            value = {}

        self._reactions = value

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def atoms(self) -> Dict[str, int]:

        all_elements = findall(chemical_formula_re, self.formula)

        atoms = {}
        for atom, count in all_elements:

            if not count:
                count = '1'

            atoms[atom] = atoms.get(atom, 0) + int(count)

        return atoms

    @property
    def molecular_weight(self) -> Union[float, int]:

        return sum([atomic_weights[atom] * count for atom, count in self.atoms.items()])

    @property
    def exchange_reaction(self) -> 'Reaction':

        for reaction in self.yield_reactions():
            if reaction.boundary:
                return reaction

    @property
    def exchange_reactions(self) -> Dict[str, 'Reaction']:

        exchanges = {}

        for reaction in self.yield_reactions():
            if reaction.boundary:
                exchanges[reaction.id] = reaction

        return exchanges

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_reactions(self) -> Generator['Reaction', None, None]:

        return generator(self._reactions)

    def yield_exchange_reactions(self) -> Generator['Reaction', None, None]:

        return generator(self.exchange_reactions)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def update(self,
               charge: int = None,
               compartment: str = None,
               formula: str = None,
               reactions: Dict[str, 'Reaction'] = None,
               **kwargs):

        super(Metabolite, self).update(**kwargs)

        if charge is not None:
            self.charge = charge

        if compartment is not None:
            self.compartment = compartment

        if formula is not None:
            self.formula = formula

        if reactions is not None:
            self._reactions.update(reactions)



from abc import ABCMeta, abstractmethod
from typing import Union, TYPE_CHECKING, Tuple, Dict, Set, List

from numpy import zeros

from mewpy.solvers import get_default_solver, solvers
from mewpy.solvers.solution import Solution
from mewpy.solvers.solver import Solver, VarType
from mewpy.mew.solution import ModelSolution

from .linear_containers import ConstraintContainer, VariableContainer
from .notification import Notification
from .linear_utils import LinkedList, Node, integer_coefficients

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


class LinearProblemInterface(metaclass=ABCMeta):
    """

    Interface for a Linear Problem. The following attributes and methods must be implemented to set up alternative
    problems

    """

    @property
    @abstractmethod
    def model(self):
        return

    @property
    @abstractmethod
    def constraints(self):
        return

    @property
    @abstractmethod
    def variables(self):
        return

    @property
    @abstractmethod
    def objective(self):
        return

    @property
    @abstractmethod
    def minimize(self):
        return

    @property
    @abstractmethod
    def solver(self):
        return

    @abstractmethod
    def notification(self, notification: Notification):
        pass

    @abstractmethod
    def optimize(self):
        pass

    @abstractmethod
    def build(self):
        pass

    @abstractmethod
    def clean(self):
        pass

    @abstractmethod
    def add(self, items):
        pass

    @abstractmethod
    def remove(self, items):
        pass

    @abstractmethod
    def update(self):
        pass


# TODO: missing documentation and typing
class LinearProblem(LinearProblemInterface):

    def __init__(self,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                 solver: Union[str, Solver] = None,
                 build: bool = True,
                 attach: bool = False):
        """

        Linear programing base implementation. A mewpy model is converted into a linear problem using reframed/mewpy
        solver interface. Both CPLEX and Gurobi solvers are currently supported. Other solvers may also be supported
        using an additional OptLang solver interface. However, CPLEX and Gurobi are recommended for certain problems.

        A linear problem is linked with a given model via asynchronous updates.
        That is, alterations to the model are sent to all attached simulators via notification objects.
        Notifications are processed accordingly by all linear problems attached to the model.
        Each implementation of a linear problem (e.g. FBA, RFBA, SRFBA, etc) is responsible
        for processing the notifications in the correct way.

        A linear problem has one and only one solver object that connect with each other via synchronous updates.
        That is, alterations to a linear problem are directly reflected in the solver.
        Alternatively, one can impose temporary constraints during problem optimization
        (see the method for further details)

        :param model: a mewpy Model, MetabolicModel, RegulatoryModel or all. The model is used to retrieve
        variables and constraints to the linear problem

        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used but it will be overwritten
        if build is true.

        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False

        """
        if not model:
            raise ValueError('A valid model must be provided')

        self._model = model

        # this is usefull to restore the linear problem to the point of init
        self._initial_solver = solver

        self._solver = self.mewpy_solver(solver)

        # Simulator index engine uses a linked list with a built-in dictionary. This allows fast access to the index
        # of a given variable or constraint.
        # Note that, some variables or constraints can comprise multiple rows or columns,
        # so that the job of keeping in track of all indexes of a given variable/constraint is actually way
        # harder than it seems for simple linear problems (e.g. fba)
        self._cols = LinkedList()
        self._rows = LinkedList()

        # one to one indexing of all variables
        self._sub_cols = LinkedList()

        # Holding constraints and variables objects
        self._constraints = {}
        self._variables = {}

        # the objective is a dict variable_id: coefficient
        self._linear_objective = {}
        self._quadratic_objective = {}
        self._minimize = True

        # Observer pattern: a linear problem object is an observer of a mewpy model.
        # When the model is changed, a notification with the change is sent to all observers (linear problems).
        # The linear problem implementation specific for each method processes the notification accordingly.
        # Finally, when the linear problem is updated,
        # all variables and constraints added to the linear problem are implemented and kept in sync with the solver
        # This can avoid consecutive building of the solver, namely a lazy loading

        # a dictionary may be too much to store pending changes, but we only want to hold the latest alterations.
        # Thus, old pending constraints or variables are replaced by the recent ones.
        # This allows the last optimization to be a direct reflection of the model latest state
        # Note that, adding and removing items from a set can be an overhead
        self._variables_queue: Dict[str, VariableContainer] = {}
        self._constraints_queue: Dict[str, ConstraintContainer] = {}

        if build:
            self.build()

        if attach:
            self.model.attach(self)

    # -----------------------------------------------------------------------------
    # Correct solver helper
    # -----------------------------------------------------------------------------
    @staticmethod
    def mewpy_solver(solver):

        if solver is None:
            solver_name = get_default_solver()

            SolverType = solvers[solver_name]

            solver = SolverType()

        elif isinstance(solver, str):

            SolverType = solvers.get(solver, None)

            if SolverType is None:
                raise ValueError(f'{solver} is not listed as valid solver. Check the valid solvers: {solvers}')

            solver = SolverType()

        elif isinstance(solver, Solver):

            pass

        else:
            raise ValueError(f'Invalid solver {solver}. Check the valid solvers: {solvers}')

        return solver

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):
        return f"{self.method} for {self.model.id}"

    def __repr__(self):
        return f"{self.method}: {self.model.id}"

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @property
    def method(self) -> str:
        """
        Name of the method implementation to build and solve the linear problem

        :return: the name of the class
        """

        return self.__class__.__name__

    @property
    def model(self) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        """

        mewpy model of this simulator

        :return: a mewpy Model, MetabolicModel, RegulatoryModel or all
        """

        return self._model

    @property
    def solver(self) -> Union[Solver]:
        """
        mewpy solver instance for this linear problem. It contains an interface for the concrete solver

        :return: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance
        """

        return self._solver

    @property
    def constraints(self) -> Dict[str, ConstraintContainer]:
        """
        A copy of the constraints container.
        This container holds all ConstraintContainer objects for this linear problem.
        Note that, a constraint container can hold several constraints/rows

        :return: copy of the constraints dictionary
        """

        return self._constraints.copy()

    @property
    def variables(self) -> Dict[str, VariableContainer]:
        """
        A copy of the variables container.
        This container holds all VariableContainer objects for this linear problem.
        Note that, a variable container can hold several variables/columns

        :return: copy of the variables dictionary
        """

        return self._variables.copy()

    @property
    def objective(self) -> Dict[Union[str, Tuple[str, str]], Union[float, int]]:
        """
        A copy of the objective dictionary. Keys are either variable identifiers or tuple of variable identifiers.
        Values are the corresponding coefficients
        Note that, linear and quadratic objectives can be encoded in the objective dictionary.
        See the set_objective method for further detail

        :return: copy of the objective dictionary
        """

        return {**self._linear_objective, **self._quadratic_objective}

    @property
    def minimize(self) -> bool:
        """

        The linear problem objective sense/direction

        :return: a boolean whether the linear problem objective sense/direction is minimization
        """

        return bool(self._minimize)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def matrix(self):
        return self._get_matrix()

    @property
    def bounds(self):
        return self.get_bounds(as_list=True)

    @property
    def b_bounds(self):
        return self.get_bounds(b_bounds=True, as_list=True)

    @property
    def shape(self):
        return int(len(self._rows)), int(len(self._cols))

    # -----------------------------------------------------------------------------
    # Observer pattern - Model notification system
    # -----------------------------------------------------------------------------

    def notification(self, notification: Notification):

        if notification.content_type == 'coefficients' and notification.action == 'set':

            variable = notification.content.id
            lb, ub = notification.content.bounds

            return self.set_bounds(variable=variable, lb=lb, ub=ub)

        elif notification.content_type == 'objectives' and notification.action == 'set':

            return self.set_objective(**notification.content)

        return

    # -----------------------------------------------------------------------------
    # Objective
    # -----------------------------------------------------------------------------

    def _set_objective(self,
                       linear: Union[str, Dict[str, Union[float, int]]] = None,
                       quadratic: Dict[Tuple[str, str], Union[float, int]] = None,
                       minimize: bool = True):
        """

        INTERNAL USE ONLY. See set_objective method

        :param linear:
        :param quadratic:
        :param minimize:
        :return:
        """

        # This helper method verify inputs and sets the objective in the linear problem object

        linear_objective = {}

        if linear:

            if isinstance(linear, str):

                linear = {linear: 1.0}

            else:

                for k, _ in linear.items():

                    if k not in self._cols and k not in self._sub_cols:
                        raise ValueError(f'{k} is not a variable of this linear problem')

            linear_objective = linear

        if linear_objective:
            self._linear_objective = linear_objective

        quadratic_objective = {}

        if quadratic:

            for (k1, k2), _ in quadratic.items():

                if k1 not in self._cols and k1 not in self._sub_cols:
                    raise ValueError(f'{k1} is not a variable of this linear problem')

                if k2 not in self._cols and k2 not in self._sub_cols:
                    raise ValueError(f'{k2} is not a variable of this linear problem')

            quadratic_objective = quadratic

        if quadratic_objective:
            self._quadratic_objective = quadratic_objective

        if minimize is not None and minimize != self._minimize:
            self._minimize = minimize

    def set_objective(self,
                      linear: Union[str, Dict[str, Union[float, int]]] = None,
                      quadratic: Dict[Tuple[str, str], Union[float, int]] = None,
                      minimize: bool = True):
        """
        A dictionary of the objective for the linear problem.
        Keys must be variables of the linear problem,
        whereas values must be the corresponding coefficients as int or float.
        It can be changed during optimization

        :param linear: a dictionary of linear coefficients or variable identifier (that is set with a coefficient of 1)
        :param quadratic: a dictionary of quadratic coefficients.
        Note that keys must be a tuple of reaction pairs to be summed up to a quadratic objective function
        :param minimize: whether to solve a minimization problem. This parameter is True by default

        :return:
        """

        # This helper method verify inputs and sets the objective in the linear problem object
        self._set_objective(linear=linear, quadratic=quadratic, minimize=minimize)

        # If the inputs are correct, the objective is set in the solver
        self._solver.set_objective(linear=linear, quadratic=quadratic, minimize=minimize)

    def set_bounds(self, variable: str, lb: Union[float, int], ub: Union[float, int]):
        """
        Set the bounds of a given linear variable available in the linear problem.
        Note that, only the bounds of atom variables can be set

        :param variable: the variable identifier
        :param lb: lower bound value
        :param ub: upper bound value

        :return:
        """

        # Either the variable is not available in the model or it is not a variable of this lp.
        # The later can occur if a multi type model has for instance an fba simulator attached
        # but a the bounds of a regulator are changed
        if variable not in self._cols and variable not in self._sub_cols:
            return

        old_lb, old_ub = self._get_variable_bounds(variable=variable)

        if len(old_lb) > 1:
            raise ValueError(f'{variable} bounds cannot be set')

        else:
            old_lb = old_lb[0]
            old_ub = old_ub[0]

        lb = lb if lb is not None else old_lb
        ub = ub if ub is not None else old_ub

        if (lb, ub) in integer_coefficients:
            var_type = VarType.INTEGER

        else:
            var_type = VarType.CONTINUOUS

        var = self._variables[variable]
        var.lbs = [lb]
        var.ubs = [ub]
        var.variables_type = [var_type]

        self.add([var])

    # -----------------------------------------------------------------------------
    # Optimization
    # -----------------------------------------------------------------------------

    def optimize(self, *args, **kwargs) -> Union[ModelSolution, Solution]:
        """

        Abstract implementation

        :return: ModelSolution or Solution (from solver) objects
        """

        # The concrete implementation is defined by each simulation method, e.g. fba, pfba, etc

    # -----------------------------------------------------------------------------
    # Operations/Manipulations - build and clean
    # -----------------------------------------------------------------------------

    @abstractmethod
    def build(self):
        """

        Abstract implementation

        :return:
        """

        # The concrete implementation is defined by each simulation method, e.g. fba, pfba, etc

    def build_solver(self):
        """

        It creates an new solver instance and adds the current state (variables, constraints) of the linear problem

        :return:
        """

        self._solver = self.mewpy_solver(self._initial_solver)

        for variable in self._variables.values():

            # Using mewpy/reframed solver interface ...
            for name, (lb, ub, var_type) in variable.items():
                self._solver.add_variable(var_id=name, lb=lb, ub=ub, vartype=var_type, update=False)

        self.solver.update()

        for i, constraint in enumerate(self._constraints.values()):

            # Using mewpy/reframed solver interface ...
            for j, (coef, lb, ub) in constraint.items():

                cnt_id = str(i + j)

                if lb == ub:
                    rhs = lb
                    self._solver.add_constraint(constr_id=cnt_id, lhs=coef, sense='=', rhs=rhs, update=False)

                else:
                    cnt_id_f = f'{cnt_id}_forward'
                    rhs = lb
                    self._solver.add_constraint(constr_id=cnt_id_f, lhs=coef, sense='>', rhs=rhs, update=False)

                    cnt_id_r = f'{cnt_id}_reverse'
                    rhs = ub
                    self._solver.add_constraint(constr_id=cnt_id_r, lhs=coef, sense='<', rhs=rhs, update=False)

        self.solver.update()

        self.set_objective(linear=self._linear_objective,
                           quadratic=self._quadratic_objective,
                           minimize=self._minimize)

    def clean(self):

        self._solver = self.mewpy_solver(self._initial_solver)
        self._cols = LinkedList()
        self._rows = LinkedList()
        self._sub_cols = LinkedList()
        self._constraints = {}
        self._variables = {}
        self._linear_objective = {}
        self._quadratic_objective = {}
        self._minimize = True
        self._variables_queue: Dict[str, VariableContainer] = {}
        self._constraints_queue: Dict[str, ConstraintContainer] = {}

    # -----------------------------------------------------------------------------
    # Operations/Manipulations - add, remove
    # -----------------------------------------------------------------------------

    def stack_container(self, container, addition):

        if isinstance(container, ConstraintContainer):
            self._constraints_queue[container.name] = (container, addition)

        if isinstance(container, VariableContainer):
            self._variables_queue[container.name] = (container, addition)

    def add(self, containers: Union[List[Union[VariableContainer, ConstraintContainer]],
                                    Set[Union[VariableContainer, ConstraintContainer]],
                                    Tuple[Union[VariableContainer, ConstraintContainer]]]):

        for container in containers:
            self.stack_container(container, addition=True)

    def remove(self, containers: Union[List[Union[VariableContainer, ConstraintContainer]],
                                       Set[Union[VariableContainer, ConstraintContainer]],
                                       Tuple[Union[VariableContainer, ConstraintContainer]]]):

        for container in containers:
            self.stack_container(container, addition=False)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations - Update
    # -----------------------------------------------------------------------------

    def update(self):

        build_problem = False

        if self._constraints_queue or self._variables_queue:
            build_problem = True

        build_new_objective = False

        for container, addition in self._variables_queue.values():

            if addition:
                self._add_variable(container)

            else:

                build_new_objective = True

                self._remove_variable(container)

        for container, addition in self._constraints_queue.values():

            if addition:
                self._add_constraint(container)

            else:
                self._remove_constraint(container)

        self._variables_queue: Dict[str, VariableContainer] = {}
        self._constraints_queue: Dict[str, ConstraintContainer] = {}

        if build_new_objective:

            new_linear = {}
            new_quadratic = {}

            for k, val in self._linear_objective.items():

                if k in self._cols or k in self._sub_cols:
                    new_linear[k] = val

            for (k1, k2), val in self._quadratic_objective.items():

                if (k1 in self._cols or k1 in self._sub_cols) and (k2 in self._cols or k2 in self._sub_cols):
                    new_quadratic[(k1, k2)] = val

            self.set_objective(linear=new_linear, quadratic=new_quadratic, minimize=self.minimize)

        # updating the solver with new constraints is horrible!!! It sometimes yields infeasible solutions
        if build_problem:
            self.build_solver()

    # -----------------------------------------------------------------------------
    # Operations/Manipulations - LinearProblem getters
    # -----------------------------------------------------------------------------

    def index(self, variable=None, constraint=None, as_list=False, as_int=False, default=None):

        if variable is None and constraint is None:
            raise ValueError('Please provide a variable or constraint')

        if constraint is not None:

            slc = self._rows.get(constraint)

        else:

            slc = self._cols.get(variable, self._sub_cols.get(variable))

            if slc is None:
                return default

        if as_list:
            return [i for i in range(slc.start, slc.stop)]

        elif as_int:
            return slc.stop - 1

        else:

            return slc

    def _get_matrix(self):

        matrix = zeros(self.shape)

        n = 0
        for cnt in self._constraints.values():

            for coef in cnt.coefs:

                for var, value in coef.items():

                    m = self.index(variable=var, as_int=True)

                    if m is None:
                        continue

                    matrix[n, m] = value

                n += 1

        return matrix

    def _get_b_bounds(self, as_list=False, as_tuples=False):

        if as_list:

            b_bounds = []

            for cnt in self._constraints.values():
                bds = list(zip(cnt.lbs, cnt.ubs))

                b_bounds.extend(bds)

            return b_bounds

        elif as_tuples:

            lbs = []
            ubs = []

            for cnt in self._constraints.values():
                lbs.extend(cnt.lbs)
                ubs.extend(cnt.ubs)

            return tuple(lbs), tuple(ubs)

        else:
            return {key: (cnt.lbs, cnt.ubs) for key, cnt in self._constraints.items()}

    def _get_bounds(self, as_list=False, as_tuples=False):

        if as_list:

            bounds = []

            for var in self._variables.values():
                bds = list(zip(var.lbs, var.ubs))

                bounds.extend(bds)

            return bounds

        elif as_tuples:

            lbs = []
            ubs = []

            for var in self._variables.values():
                lbs.extend(var.lbs)
                ubs.extend(var.ubs)

            return tuple(lbs), tuple(ubs)

        else:
            return {key: (var.lbs, var.ubs) for key, var in self._variables.items()}

    def _get_variable_bounds(self, variable):

        variable = self._variables.get(variable)

        if variable is None:
            lb, ub = self._get_bounds(as_list=True)[self._sub_cols.get(variable)]

            return [lb], [ub]

        return variable.lbs, variable.ubs

    def _get_constraint_bounds(self, constraint):

        constraint = self._constraints.get(constraint)

        return constraint.lbs, constraint.ubs

    def get_bounds(self,
                   variable=None,
                   constraint=None,
                   b_bounds=False,
                   as_list=False,
                   as_tuples=False
                   ):

        if variable is not None:

            return self._get_variable_bounds(variable=variable)

        elif constraint is not None:

            return self._get_constraint_bounds(constraint)

        elif b_bounds:

            return self._get_b_bounds(as_list=as_list, as_tuples=as_tuples)

        else:
            return self._get_bounds(as_list=as_list, as_tuples=as_tuples)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations - LinearProblem add/remove variables and constraints
    # -----------------------------------------------------------------------------

    def _update_constraint_coefs(self, constraint: ConstraintContainer):

        # some constraint coefficients might have keys that refer to the variable name and not sub-variable name.
        # Since only sub-variable names are added to the solvers,
        # these keys must be updated to the last sub-variable name. Note that, the last sub-variable name is regularly
        # the one that matters, as the initial sub-variables regularly decide the outcome of the last one.

        new_coefs = []

        for coefficient in constraint:

            new_coef = {}

            for var, coef in coefficient.items():

                if var in self._sub_cols:

                    sub_var = var

                else:

                    variable = self._variables[var]

                    sub_var = variable.sub_variables[-1]

                new_coef[sub_var] = coef

            new_coefs.append(new_coef)

        constraint.coefs = new_coefs

    def _add_constraint(self, constraint: ConstraintContainer):

        if constraint.name in self._rows:
            # The constraint is replaced, as the linear problem behaves like a set
            # This also mimics the solver interface behavior

            old_constraint = self._constraints[constraint.name]

            self._remove_constraint(old_constraint)

        node = constraint.to_node()

        self._rows.add(node)

        self._update_constraint_coefs(constraint)

        self._constraints[constraint.name] = constraint

    def _remove_constraint(self, constraint: ConstraintContainer):

        if constraint.name in self._rows:
            self._rows.pop(constraint.name)
            self._constraints.pop(constraint.name)

    def _add_variable(self, variable: VariableContainer):

        if variable.name in self._cols:
            # The variable is replaced, as the linear problem behaves like a set
            # This also mimics the solver interface behavior

            old_variable = self._variables[variable.name]

            self._remove_variable(old_variable)

        node = variable.to_node()

        self._cols.add(node)

        self._variables[variable.name] = variable

        for sub_variable in variable.keys():
            sub_node = Node(value=sub_variable, length=1)

            self._sub_cols.add(sub_node)

    def _remove_variable(self, variable: VariableContainer):

        if variable.name in self._cols:

            self._cols.pop(variable.name)
            self._variables.pop(variable.name)

            for sub_variable in variable.keys():
                self._sub_cols.pop(sub_variable)



from typing import Union, TYPE_CHECKING

from mewpy.simulation import get_container, get_simulator

from .director import Director
from .reader import Reader
from .writer import Writer

from .engines import (Engines,
                      RegulatoryCSV,
                      CoExpressionCSV,
                      TargetRegulatorCSV,
                      CobrapyModel,
                      ReframedModel,
                      JSON,
                      RegulatorySBML,
                      MetabolicSBML)
from .engines.csv import read_gene_expression_dataset, read_coregflux_influence_matrix

if TYPE_CHECKING:
    from io import TextIOWrapper

    from mewpy.model import Model, RegulatoryModel, MetabolicModel

    try:
        # noinspection PyPackageRequirements
        from cobra import Model as Cobra_Model

    except ImportError:
        Cobra_Model = None

    try:
        # noinspection PyPackageRequirements
        from reframed import CBModel as Reframed_Model

    except ImportError:

        Reframed_Model = None


# TODO: NOTE: The following load functions were in the sbml module.
#  I moved them here, so that the sbml file can be removed and thus not being confused with the sbml module under
#  the engines sub-package
def load_sbml_container(filename, flavor='reframed'):
    if flavor == 'reframed':
        from reframed.io.sbml import load_cbmodel
        model = load_cbmodel(filename)
    elif flavor == 'cobra':
        from cobra.io import read_sbml_model
        model = read_sbml_model(filename)
    else:
        raise ValueError(f"{flavor} is not a recognized flavor")
    container = get_container(model)
    return container


def load_sbml_simulator(filename, flavor='reframed', envcond=None):
    if flavor == 'reframed':
        from reframed.io.sbml import load_cbmodel
        model = load_cbmodel(filename)
    elif flavor == 'cobra':
        from cobra.io import read_sbml_model
        model = read_sbml_model(filename)
    else:
        raise ValueError(f"{flavor} is not a recognized flavor")
    simul = get_simulator(model, envcond=envcond)
    return simul


def load_gecko_simulator(filename, flavor='reframed', envcond=None):
    if flavor == 'reframed':
        from mewpy.model.gecko import GeckoModel
        model = GeckoModel(filename)
    elif flavor == 'cobra':
        from geckopy.gecko import GeckoModel
        model = GeckoModel(filename)
    else:
        raise ValueError(f"{flavor} is not a recognized flavor")
    simul = get_simulator(model, envcond=envcond)
    return simul


# A common entry point for all models.
# The following "read" and "write" functions available in this module are just wrappers for a single reader/writer
# or multiple readers/writers using the director
def read_model(*readers: Reader,
               warnings: bool = True) -> Union['Model', 'RegulatoryModel', 'MetabolicModel']:

    """
    Reading a mewpy model encoded into one or more file types (e.g. sbml, csv, cobrapy, reframed, json, etc).
    It can return a metabolic, regulatory or metabolic-regulatory model from multiple files.

    A reader must be provided for each file type.
    Reading will take place according to the order and settings of the each reader.

    All files are closed upon reading or failure

    :param readers: Multiple Reader instances that will be used to read multiple file types into a single mewpy model
    :param warnings: Whether to launch warnings found during reading
    :return: mewpy metabolic, regulatory or both model
    """

    reader_director = Director(*readers)

    model = reader_director.read()

    if warnings:
        reader_director.warn()

    return model


def read_sbml(io: Union[str, 'TextIOWrapper'],
              metabolic: bool = True,
              regulatory: bool = True,
              warnings: bool = True) -> Union['Model', 'RegulatoryModel', 'MetabolicModel']:

    """
    Reading a mewpy model encoded into a SBML file.
    It can return a metabolic, regulatory or metabolic-regulatory model from the SBML file according to the metabolic
    and regulatory flags.

    Only one SBML file is accepted. Thus, if the SBML file does not encode an SBML-qual or SBML-fbc plugin,
    but either metabolic or regulatory reading are requested, an IO error will be raised, respectively.

    The SBML file is closed upon reading or failure

    :param io: A valid string path or IO.
    :param metabolic: Whether to read the metabolic share of the SBML file, namely the fbc plugin
    :param regulatory: Whether to read the regulatory share of the SBML file, namely the qual plugin
    :param warnings: Whether to launch warnings found during reading
    :return: mewpy metabolic, regulatory or both model
    """

    readers = []

    if metabolic:
        metabolic_reader = Reader(engine=MetabolicSBML,
                                  io=io)

        readers.append(metabolic_reader)

    if regulatory:
        regulatory_reader = Reader(engine=RegulatorySBML,
                                   io=io)

        readers.append(regulatory_reader)

    if not readers:
        raise OSError('Nothing to read')

    return read_model(*readers, warnings=warnings)


def read_csv(io: Union[str, 'TextIOWrapper'],
             boolean: bool = True,
             co_expression: bool = False,
             target_regulator: bool = False,
             warnings: bool = True,
             **kwargs) -> Union['Model', 'RegulatoryModel']:

    """
    Reading a mewpy regulatory model encoded into a CSV file.
    It can only return a regulatory model from the CSV file.

    Only one CSV file is accepted. The CSV file type can be explicitly set using the boolean (default), co_expression
    or target_regulator flags. Consult the Engines enumerator in mewpy.io.engines for further detail on this CSV file
    types.

    The CSV file is closed upon reading or failure

    :param io: A valid string path or IO.
    :param boolean: Whether the file is a Boolean-based regulatory CSV file
    :param co_expression: Whether the file is a CoExpression (co-activating and co-repressing) regulatory CSV file
    :param target_regulator: Whether the file is a Target-Regulator interaction-based regulatory CSV file
    :param warnings: Whether to launch warnings found during reading
    :return: mewpy regulatory model
    """

    readers = []

    if boolean:
        boolean_reader = Reader(engine=RegulatoryCSV,
                                io=io,
                                **kwargs)

        readers.append(boolean_reader)

    if co_expression:
        regulatory_reader = Reader(engine=CoExpressionCSV,
                                   io=io,
                                   **kwargs)

        readers.append(regulatory_reader)

    if target_regulator:
        regulatory_reader = Reader(engine=TargetRegulatorCSV,
                                   io=io,
                                   **kwargs)

        readers.append(regulatory_reader)

    if not readers:
        raise OSError('Nothing to read')

    if len(readers) > 1:
        raise OSError('read_csv only accepts one file at a time')

    return read_model(*readers, warnings=warnings)


def read_cbmodel(io: Union['Cobra_Model', 'Reframed_Model'],
                 cobrapy: bool = True,
                 reframed: bool = False,
                 warnings: bool = True) -> Union['Model', 'MetabolicModel']:

    """
    Reading a mewpy metabolic model encoded into a Constraint-Based metabolic model from Cobrapy or Reframed.
    It can only return a metabolic model from the cobra model.

    Only one cobra model. The cobra model platform can be explicitly set using the cobrapy (default) or reframed flags.
    Consult the Engines enumerator in mewpy.io.engines for further detail on this cobra model types.

    :param io: A valid cobra model.
    :param cobrapy: Whether the cobra model is a cobrapy model
    :param reframed: Whether the cobra model is a reframed model
    :param warnings: Whether to launch warnings found during reading
    :return: mewpy metabolic model
    """

    readers = []

    if cobrapy:
        boolean_reader = Reader(engine=CobrapyModel,
                                io=io)

        readers.append(boolean_reader)

    if reframed:
        regulatory_reader = Reader(engine=ReframedModel,
                                   io=io)

        readers.append(regulatory_reader)

    if not readers:
        raise OSError('Nothing to read')

    if len(readers) > 1:
        raise OSError('read_cbmodel only accepts one cbmodel at a time')

    return read_model(*readers, warnings=warnings)


def read_json(io: Union[str, 'TextIOWrapper'],
              warnings: bool = True) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:

    """
    Reading a mewpy model encoded into a JSON file.
    It can return a metabolic, regulatory or metabolic-regulatory model from the JSON file according to the JSON file.

    Only one JSON file is accepted. The mewpy model is built according to the JSON file content.

    The JSON file is closed upon reading or failure

    :param io: A valid string path or IO.
    :param warnings: Whether to launch warnings found during reading
    :return: mewpy metabolic, regulatory or both model
    """

    json_reader = Reader(engine=JSON,
                         io=io)

    return read_model(json_reader, warnings=warnings)


def write_model(*writers: Writer,
                warnings=True) -> Union['Model', 'RegulatoryModel', 'MetabolicModel']:

    """
    Writing a mewpy model into one or more file types (e.g. sbml, csv, cobrapy, reframed, json, etc).
    It can write a metabolic, regulatory or metabolic-regulatory model to multiple files.

    A writer must be provided for each file type.
    Reading will take place according to the order and settings of the each writer.

    All files are closed upon writing or failure

    :param writers: Multiple Writer instances that will be used to write multiple file types from a single mewpy model
    :param warnings: Whether to launch warnings found during reading
    :return: mewpy metabolic, regulatory or both model
    """

    writer_director = Director(*writers)

    model = writer_director.write()

    if warnings:
        writer_director.warn()

    return model


from itertools import product
from typing import Dict, Union, TYPE_CHECKING, Callable

# TODO: this module depends on pandas dataframes. Should it be set as package requirement?
# noinspection PyPackageRequirements
from pandas import DataFrame

from .algebra_utils import solution_decode
from .parsing import tokenize
from .symbolic import _walk, NoneAtom, Symbolic

if TYPE_CHECKING:
    from mewpy.mew.variables import Gene, Metabolite, Reaction, Regulator


# TODO: methods stubs and type hinting
class Expression:

    def __init__(self,
                 symbolic: Symbolic = None,
                 variables: Dict[str, Union['Gene', 'Metabolite', 'Reaction', 'Regulator']] = None):
        """
        Association between a symbolic mathematical expression and variables.

        :param symbolic:
        :param variables:

        """

        if symbolic is None:
            symbolic = NoneAtom()

        if variables is None:
            variables = {}

        self._symbolic: Symbolic = symbolic
        self._variables: Dict[str, Union['Gene', 'Metabolite', 'Reaction', 'Regulator']] = variables

        self.check_association()

    def check_association(self):

        for symbol in self.symbols.values():

            if symbol.name not in self._variables:
                raise ValueError('Expression set with incorrect variables or symbolic expression')

    @property
    def symbolic(self) -> Symbolic:
        return self._symbolic

    @property
    def variables(self) -> Dict[str, Union['Gene', 'Metabolite', 'Reaction', 'Regulator']]:
        return self._variables.copy()

    @variables.setter
    def variables(self, value):

        self._variables = value

        self.check_association()

    @property
    def symbols(self):
        return {symbol.name: symbol for symbol in self.symbolic.atoms(symbols_only=True)}

    @property
    def is_none(self):
        return self.symbolic.is_none

    def __repr__(self):
        return repr(self.symbolic)

    def __str__(self):
        return str(self.symbolic)

    def to_string(self):
        return self.symbolic.to_string()

    def to_tokenize(self):
        return tokenize(self.to_string())

    def __iter__(self):

        return self.symbolic.__iter__()

    def __next__(self):

        return self.symbolic.__next__()

    def walk(self, reverse=False):

        return _walk(self.symbolic, reverse)

    def __call__(self,
                 values: Dict[str, int],
                 coefficient: Union[float, int] = None,
                 operators: Dict[Symbolic, Callable] = None,
                 default: Union[float, int] = 0.0,
                 decoder: dict = None,
                 **kwargs):

        return self.evaluate(values=values,
                             coefficient=coefficient,
                             operators=operators,
                             default=default,
                             decoder=decoder,
                             **kwargs)

    def evaluate(self,
                 values: Dict[str, int],
                 coefficient: Union[float, int] = None,
                 operators: Dict[Symbolic, Callable] = None,
                 default: Union[float, int] = 0.0,
                 decoder: dict = None,
                 **kwargs):

        res = self.symbolic.evaluate(values=values, operators=operators, default=default, **kwargs)

        res = solution_decode(res, decoder)

        # if a coefficient is provided and the expression is evaluated to true, the respective coefficient will
        # be returned
        if coefficient is None:
            return res

        else:

            if res:
                return coefficient

            return res

    def truth_table(self,
                    values: Dict[str, Union[float, int]] = None,
                    active_states: bool = True,
                    coefficient: Union[float, int] = None,
                    operators: Dict[Symbolic, Callable] = None,
                    default: Union[float, int] = 0.0,
                    decoder: dict = None):

        truth_table = []

        if values:

            values['result'] = self.evaluate(values=values,
                                             coefficient=coefficient,
                                             operators=operators,
                                             default=default,
                                             decoder=decoder)

            truth_table.append(values)

        else:
            if active_states:
                values = {key: variable.coefficient.active_coefficient
                          for key, variable in self._variables.items()}

                values['result'] = self.evaluate(values=values,
                                                 coefficient=coefficient,
                                                 operators=operators,
                                                 default=default,
                                                 decoder=decoder)

                truth_table.append(values)

            else:

                variables = list(self.variables.keys())
                coefficients = [variable.coefficient.coefficients for variable in self._variables.values()]

                for combined_coefficients in product(*coefficients):
                    values = dict(list(zip(variables, combined_coefficients)))

                    values['result'] = self.evaluate(values=values,
                                                     coefficient=coefficient,
                                                     operators=operators,
                                                     default=default,
                                                     decoder=decoder)

                    truth_table.append(values)

        return DataFrame(truth_table)



from dataclasses import dataclass, field
from typing import Any, Dict, Union, TYPE_CHECKING, Tuple, Set

from pandas import DataFrame

try:
    # noinspection PyPackageRequirements
    from cobra import Model as Cobra_Model

except ImportError:
    Cobra_Model = str

try:
    # noinspection PyPackageRequirements
    from reframed import CBModel as Reframed_Model

except ImportError:

    Reframed_Model = str

from mewpy.mew.algebra import Symbolic, NoneAtom
from mewpy.mew.variables import Variable

if TYPE_CHECKING:
    from mewpy.mew.variables import Gene, Interaction, Metabolite, Reaction, Regulator, Target
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


@dataclass
class History:

    """
    Data transfer object for history.
    History encoded into a sbml model
    """

    data: str = None
    creators: str = None


@dataclass
class FunctionTerm:

    """
    Data transfer object for function terms.
    Function term holds the symbolic expression of a given interaction between a target and a set of regulators.
    It also holds the resulting coefficient of this interaction
    """

    id: str = None
    symbolic: Symbolic = field(default_factory=NoneAtom)
    coefficient: int = 0


@dataclass
class CompartmentRecord:

    """
    Data transfer object for compartments.
    """

    id: str = 'e'
    name: str = 'external'


@dataclass
class VariableRecord:

    """
    Data transfer object for variables encoded into files.
    Variable record holds information regarding any multi-type variable, namely metabolic or regulatory.
    It can be extended to more types.
    It stores the many attributes that these variables can have encoded into multiple file types.
    """

    # ids, names, types, etc
    id: Any = field(default_factory=str)
    name: str = field(default_factory=str)
    aliases: set = field(default_factory=set)

    # other
    notes: Any = None
    annotation: Any = None

    # common attributes
    compartment: Any = None
    coefficients: set = field(default_factory=set)
    active_coefficient: float = 0.0
    constant: Any = None

    # metabolic attributes
    bounds: tuple = None
    charge: int = None
    formula: str = None
    genes: Dict[str, 'VariableRecord'] = field(default_factory=dict)
    gpr: FunctionTerm = field(default_factory=FunctionTerm)
    metabolites: Dict[str, 'VariableRecord'] = field(default_factory=dict)
    products: Dict[str, 'VariableRecord'] = field(default_factory=dict)
    reactants: Dict[str, 'VariableRecord'] = field(default_factory=dict)
    stoichiometry: Dict[str, Union[int, float]] = field(default_factory=dict)

    # regulatory attributes
    regulators: Dict[str, 'VariableRecord'] = field(default_factory=dict)
    target: 'VariableRecord' = None
    interactions: Dict[str, 'VariableRecord'] = field(default_factory=dict)
    function_terms: Dict[str, 'FunctionTerm'] = field(default_factory=dict)

    # TODO: it should be improved for further automation, so that one can just type to_variable(). For now,
    #  one still has to pass the correct attributes at each reading/model building
    def to_variable(self,
                    model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                    types: Set[str],
                    **attributes) -> Tuple[Union['Gene', 'Interaction', 'Metabolite', 'Reaction', 'Regulator',
                                                 'Target'], str]:

        try:

            variable = model.get(self.id, None)

        except AttributeError:

            warning = f'{self.id} cannot be built properly. Generic Variable, build instead'

            return Variable(identifier=self.id), warning

        if variable is None:

            try:

                attributes['identifier'] = self.id

                variable = Variable.from_types(types=types, **attributes)

                return variable, ''

            except TypeError:

                warning = f'{self.id} cannot be built properly. Generic Variable, build instead'

                return Variable(identifier=self.id), warning

        else:

            try:

                variable.update(**attributes)

                return variable, ''

            except TypeError:

                warning = f'{self.id} cannot be built properly. Generic Variable, build instead'

                return variable, warning


@dataclass
class DataTransferObject:
    """
    Data transfer object for models encoded into files. This is the primary DTO associated with a builder.
    DTO record holds information regarding any multi-type model, namely metabolic or regulatory.
    It can be extended to more types.
    It stores the many attributes that these models can have encoded into multiple file types.
    """

    # cobra model
    cobra_model: Cobra_Model = None

    # reframed model
    reframed_model: Reframed_Model = None

    # CSV, TXT dataframe
    data_frame: DataFrame = field(default_factory=DataFrame)
    aliases_columns: list = field(default_factory=list)

    # SBML doc & model
    doc: Any = None
    model: Any = None

    # SBML plugins
    qual_plugin: Any = None
    fbc_plugin: Any = None

    # ids, names, versions, etc
    id: Any = None
    name: str = None
    level: Any = None
    version: Any = None
    history: Any = None

    # types
    types: set = None

    # common containers
    compartments: Dict[str, VariableRecord] = field(default_factory=dict)
    extracellular_reactions: Dict[str, VariableRecord] = field(default_factory=dict)
    variables: Dict[str, VariableRecord] = field(default_factory=dict)

    # metabolic containers
    extracellular_metabolites: Dict[str, VariableRecord] = field(default_factory=dict)
    genes: Dict[str, VariableRecord] = field(default_factory=dict)
    metabolites: Dict[str, VariableRecord] = field(default_factory=dict)
    objective: Dict[str, VariableRecord] = field(default_factory=dict)
    reactions: Dict[str, VariableRecord] = field(default_factory=dict)

    # regulatory containers
    interactions: Dict[str, VariableRecord] = field(default_factory=dict)
    regulators: Dict[str, VariableRecord] = field(default_factory=dict)
    targets: Dict[str, VariableRecord] = field(default_factory=dict)



from enum import Enum

from .fba import FBA, milpFBA, pFBA
from .rfba import RFBA
from .srfba import SRFBA
from .sim_bool import SimBool
from .milp_bool import milpBool

from .metabolic_analysis import slim_fba, slim_pfba, slim_milp_fba, fva, single_gene_deletion, single_reaction_deletion
from .regulatory_analysis import slim_sim_bool, slim_milp_bool, single_regulator_deletion, regulatory_events
from .integrated_analysis import (slim_rfba, slim_srfba, ifva, isingle_regulator_deletion, isingle_reaction_deletion,
                                  isingle_gene_deletion)


class Analysis(Enum):
    FBA = FBA
    milpFBA = milpFBA
    pFBA = pFBA
    RFBA = RFBA
    SRFBA = SRFBA
    SynchronousBoolean = SimBool
    BooleanMILP = milpBool

    @classmethod
    def has_analysis(cls, analysis):
        try:
            return cls[analysis]

        except KeyError:

            return False

    @classmethod
    def get(cls, analysis, default=FBA):
        try:
            return cls[analysis]

        except KeyError:

            return default



from typing import Any, TYPE_CHECKING, Set, Union, Dict, Generator, Tuple, List

from mewpy.util.utilities import generator
from mewpy.util.serialization import serialize
from mewpy.util.history import recorder
from .coefficient import Coefficient
from .variable import Variable

if TYPE_CHECKING:
    from .interaction import Interaction
    from .regulator import Regulator


# TODO: methods stubs
class Target(Variable, variable_type='target', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
                 active_coefficient: Union[int, float] = None,
                 interaction: 'Interaction' = None,
                 **kwargs):

        """
        A target is commonly associated with one and only one interaction but
        can usually be available as gene and regulator too.

        It holds information regarding the coefficients that can take and the interaction that is modelling these
        coefficients

        :param identifier: identifier, e.g. b0001
        :param coefficients: the set of coefficients that this target can take.
        These coefficients can be expanded later. 0 and 1 are added by default
        :param active_coefficient: the active coefficient
        :param interactions: the  interaction to which the target is associated with
        """

        # the coefficient initializer sets minimum and maximum coefficients of 0.0 and 1.0
        if not coefficients:
            coefficients = (0.0, 1.0)

        if not active_coefficient:
            active_coefficient = max(coefficients)

        if not interaction:
            interaction = None

        self._coefficient = Coefficient(variable=self, coefficients=coefficients, active=active_coefficient)
        self._interaction = interaction

        super().__init__(identifier,
                         **kwargs)

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Target.variable_type}

        _types.update(super(Target, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):

        if self.interaction:

            return f'{self.interaction}'

        return super(Target, self).__str__()

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('coefficient', 'coefficients', '_coefficient')
    @property
    def coefficient(self) -> Coefficient:

        if hasattr(self, '_bounds'):

            # if it is a reaction, the bounds coefficient must be returned
            return self._bounds

        elif hasattr(self, 'exchange_reaction'):

            # if it is a metabolite, the bounds coefficient of the exchange reaction must be returned
            if hasattr(self.exchange_reaction, '_bounds'):
                # noinspection PyProtectedMember
                return self.exchange_reaction._bounds

        return self._coefficient

    @serialize('interaction', 'interaction', '_interaction')
    @property
    def interaction(self) -> 'Interaction':
        return self._interaction

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @interaction.setter
    @recorder
    def interaction(self, value: 'Interaction'):

        if not value:
            value = None

        value.add_target(self, history=False)

        self._interaction = value

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def regulators(self) -> Dict[str, 'Regulator']:

        if self.interaction:
            return self.interaction.regulators.copy()

        return {}

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_regulators(self) -> Generator['Regulator', None, None]:
        return generator(self.regulators)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def ko(self, minimum_coefficient: Union[int, float] = 0.0, history=True):

        return self.coefficient.ko(minimum_coefficient=minimum_coefficient, history=history)

    def update(self,
               coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
               active_coefficient: Union[int, float] = None,
               interaction: 'Interaction' = None,
               **kwargs):

        super(Target, self).update(**kwargs)

        if coefficients is not None:
            self.coefficient.coefficients = coefficients

        if active_coefficient is not None:
            self.coefficient.active_coefficient = active_coefficient

        if interaction is not None:
            self.interaction = interaction



"""An interface for ODE solvers"""
from enum import Enum
from abc import ABC, abstractmethod


class ODEMethod(Enum):
    RK45 = 'RK45',
    RK23 = 'RK23',
    DOP853 = 'DOP853',
    Radau = 'Radau',
    BDF = 'BDF',
    LSODA = 'LSODA',
    LSODAR = 'LSODAR',
    LSODE = 'LSODE',
    HEUN = 'HEUN',
    EULER = 'EULER',
    RK4 = 'RK4',
    DORMAN_PRINCE = 'DORMAN_PRINCE',
    RKFehlberg = 'RKFehlberg',
    Dopri5 = 'Dopri5',
    Vode = 'Vode',
    Radau5 = 'Radau5',
    AdamsBashforth2 = 'AdamsBashforth2',
    AdamsBashMoulton2 = 'AdamsBashMoulton2'

    def __eq__(self, other):
        """Overrides equal to enable string name comparison"""
        if isinstance(other, ODEMethod):
            return super().__eq__(other)
        elif isinstance(other, str):
            return self.name == other
        else:
            return False

    def __hash__(self):
        return hash(self.name)


class SolverConfigurations:
    ABSOLUTE_TOL = 1e-9
    RELATIVE_TOL = 1e-6
    N_STEPS = 10000


class KineticConfigurations:
    SOLVER_METHOD = ODEMethod.Radau
    STEADY_STATE_TIME = 1e9
    SOLVER_TIMEOUT = 6000


class ODEStatus(Enum):
    ERROR = 0
    OPTIMAL = 1


class ODESolver(ABC):

    @abstractmethod
    def solve(self, y0, t_span, **kwargs):
        raise NotImplementedError



def solution_decode(solution, decoder=None):
    if not decoder:
        decoder = {True: 1,
                   False: 0,
                   1: 1,
                   0: 0,
                   -1: 1,
                   -2: 0
                   }

    return decoder.get(solution, solution)



from typing import Any, Set, Union, Dict, TYPE_CHECKING, Generator, Tuple, List

from mewpy.util.utilities import generator
from mewpy.util.serialization import serialize
from .coefficient import Coefficient
from .variable import Variable

if TYPE_CHECKING:
    from .interaction import Interaction
    from .target import Target


# TODO: methods stubs
class Regulator(Variable, variable_type='regulator', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
                 active_coefficient: Union[int, float] = None,
                 interactions: Dict[str, 'Interaction'] = None,
                 **kwargs):

        """
        A regulator is commonly associated with interactions and
        can usually be available as metabolite or reaction or target too.

        It holds information regarding the coefficients that can take and the interactions to which is associated

        :param identifier: identifier, e.g. b0001
        :param coefficients: the set of coefficients that this regulator can take.
        These coefficients can be expanded later. 0 and 1 are added by default
        :param active_coefficient: the active coefficient
        :param interactions: the dictionary of interactions to which the regulator is associated with
        """

        # the coefficient initializer sets minimum and maximum coefficients of 0.0 and 1.0
        if not coefficients:
            coefficients = (0.0, 1.0)

        if not active_coefficient:
            active_coefficient = min(coefficients)

        if not interactions:
            interactions = {}

        self._coefficient = Coefficient(variable=self, coefficients=coefficients, active=active_coefficient)
        self._interactions = interactions

        super().__init__(identifier,
                         **kwargs)

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Regulator.variable_type}

        _types.update(super(Regulator, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('coefficient', 'coefficients', '_coefficient')
    @property
    def coefficient(self) -> Coefficient:

        if hasattr(self, '_bounds'):

            # if it is a reaction, the bounds coefficient must be returned
            return self._bounds

        elif hasattr(self, 'exchange_reaction'):

            # if it is a metabolite, the bounds coefficient of the exchange reaction must be returned
            if hasattr(self.exchange_reaction, '_bounds'):
                # noinspection PyProtectedMember
                return self.exchange_reaction._bounds

        return self._coefficient

    @serialize('interactions', 'interactions', '_interactions')
    @property
    def interactions(self) -> Dict[str, 'Interaction']:
        return self._interactions.copy()

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @interactions.setter
    def interactions(self, value: Dict[str, 'Interaction']):

        if not value:
            value = {}

        self._interactions = value

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def targets(self) -> Dict[str, 'Target']:

        return {interaction.target.id: interaction.target for interaction in self.yield_interactions()
                if interaction.target is not None}

    @property
    def environmental_stimulus(self) -> bool:

        if self.types == {'regulator'}:
            return True

        elif self.is_regulator() and self.is_metabolite():
            return True

        elif self.is_regulator() and self.is_reaction():
            return True

        return False

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_interactions(self) -> Generator['Interaction', None, None]:

        return generator(self._interactions)

    def yield_targets(self) -> Generator['Target', None, None]:

        return generator(self.targets)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def ko(self, minimum_coefficient: Union[int, float] = 0.0, history=True):

        return self.coefficient.ko(minimum_coefficient=minimum_coefficient, history=history)

    def update(self,
               coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
               active_coefficient: Union[int, float] = None,
               interactions: Dict[str, 'Interaction'] = None,
               **kwargs):

        super().update(**kwargs)

        if coefficients is not None:
            self.coefficient.coefficients = coefficients

        if active_coefficient is not None:
            self.coefficient.active_coefficient = active_coefficient

        if interactions is not None:
            self._interactions.update(interactions)



from abc import abstractmethod
from functools import partial
from typing import List, TYPE_CHECKING

from mewpy.solvers.solver import VarType
from mewpy.util.constants import ModelConstants
from mewpy.util.utilities import iterable

from .linear_problem import LinearProblem
from .linear_containers import ConstraintContainer, VariableContainer, concat_constraints
from .linear_utils import integer_coefficients, bounds_from_symbol, bounds_from_variable
from .notification import Notification

if TYPE_CHECKING:
    from mewpy.mew.algebra import Symbolic
    from mewpy.mew.variables import Reaction, Metabolite

MEWPY_LB = ModelConstants.REACTION_LOWER_BOUND
MEWPY_UB = ModelConstants.REACTION_UPPER_BOUND
MEWPY_TOL = ModelConstants.TOLERANCE


# TODO: missing documentation and typing
class MetabolicLinearizer(LinearProblem):

    @abstractmethod
    def build(self):

        # The concrete implementation is defined by each simulation method, e.g. fba, pfba, etc

        pass

    def notification(self, notification: Notification):

        if notification.content_type == 'reactions' and notification.action == 'add':

            return self.add_reactions(notification.content)

        elif notification.content_type == 'reactions' and notification.action == 'remove':

            return self.remove_reactions(notification.content)

        elif notification.content_type == 'metabolites' and notification.action == 'add':

            return self.add_metabolites(notification.content)

        elif notification.content_type == 'metabolites' and notification.action == 'remove':

            return self.remove_metabolites(notification.content)

        else:
            return super(MetabolicLinearizer, self).notification(notification)

    def metabolite_reaction_lookup(self, reactions: List['Reaction'], metabolites: List['Metabolite']):

        reactions: List['Reaction'] = iterable(reactions)
        metabolites: List['Metabolite'] = iterable(metabolites)

        constraints = {metabolite.id: ConstraintContainer(name=metabolite.id,
                                                          coefs=[{}],
                                                          lbs=[0.0],
                                                          ubs=[0.0])
                       for metabolite in metabolites}

        variables = []
        for reaction in reactions:

            lb, ub = reaction.bounds

            var = VariableContainer(name=reaction.id,
                                    sub_variables=[reaction.id],
                                    lbs=[float(lb)],
                                    ubs=[float(ub)],
                                    variables_type=[VarType.CONTINUOUS])

            variables.append(var)

            for metabolite, coef in reaction.stoichiometry.items():
                metabolite_cnt = constraints[metabolite.id]
                metabolite_cnt.coefs[0][reaction.id] = coef

        self.add(variables)
        self.add(constraints.values())

    def add_reactions(self, reactions: List['Reaction']):

        reactions: List['Reaction'] = iterable(reactions)

        variables = []
        constraints = []
        for reaction in reactions:

            lb, ub = reaction.bounds

            var = VariableContainer(name=reaction.id,
                                    sub_variables=[reaction.id],
                                    lbs=[float(lb)],
                                    ubs=[float(ub)],
                                    variables_type=[VarType.CONTINUOUS])

            variables.append(var)

            for metabolite, coef in reaction.stoichiometry.items():
                constraint = self._constraints.get(metabolite.id, ConstraintContainer(name=metabolite.id,
                                                                                      coefs=[{}],
                                                                                      lbs=[0.0],
                                                                                      ubs=[0.0]))

                constraint.coefs[0][reaction.id] = coef
                constraints.append(constraint)

        self.add(variables)
        self.add(constraints)

    def remove_reactions(self, reactions: List['Reaction']):

        reactions: List['Reaction'] = iterable(reactions)

        variables = []
        constraints = []
        for reaction in reactions:

            if reaction.id in self._variables:
                variables.append(self._variables[reaction.id])

            for metabolite in reaction.metabolites:

                if metabolite in self._constraints:

                    constraint = self._constraints[metabolite]

                    if reaction.id in constraint.coefs[0]:
                        del constraint.coefs[0][reaction.id]

                    constraints.append(constraint)

        self.remove(variables)
        self.remove(constraints)

    def add_metabolites(self, metabolites: List['Metabolite']):

        metabolites: List['Metabolite'] = iterable(metabolites)

        constraints = [ConstraintContainer(name=metabolite.id,
                                           coefs=[{}],
                                           lbs=[0.0],
                                           ubs=[0.0])
                       for metabolite in metabolites]

        self.add(constraints)

    def remove_metabolites(self, metabolites: List['Metabolite']):

        metabolites: List['Metabolite'] = iterable(metabolites)

        self.remove([self._constraints[metabolite.id] for metabolite in metabolites
                     if metabolite.id in self._constraints])


class LogicLinearizer(LinearProblem):

    @abstractmethod
    def build(self):

        # The concrete implementation is defined by each simulation method, e.g. fba, pfba, etc

        pass

    def notification(self, notification: Notification):

        # The concrete implementation is defined by each simulation method, e.g. fba, pfba, etc

        return super(LogicLinearizer, self).notification(notification)

    @staticmethod
    def linearize_and(symbolic):

        # Following Boolean algebra, an And (a = b and c) can be translated as: a = b*c
        # Alternatively, an And can be written as lb < b + c - a < ub

        # So, for the mid term expression a = b and c
        # We have therefore the equation: -1 <= 2*b + 2*c – 4*a <= 3

        name = symbolic.key()
        names = [f'{name}_{i}' for i, _ in enumerate(symbolic.variables[:-1])]

        and_op = names[0]
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        _coefs = []
        _lbs = []
        _ubs = []

        if op_l.is_one or op_l.is_true:

            _coef = {and_op: -4.0, op_r.key(): 2.0}
            _state = (-3.0, 1.0)

        elif op_r.is_one or op_r.is_true:

            _coef = {and_op: -4.0, op_l.key(): 2.0}
            _state = (-3.0, 1.0)

        elif op_l.is_zero or op_l.is_false:

            _coef = {and_op: -4.0, op_r.key(): 2.0}
            _state = (-1.0, 3.0)

        elif op_r.is_zero or op_r.is_false:

            _coef = {and_op: -4.0, op_l.key(): 2.0}
            _state = (-1.0, 3.0)

        else:

            _coef = {and_op: -4.0, op_l.key(): 2.0, op_r.key(): 2.0}
            _state = (-1.0, 3.0)

        _coefs.append(_coef)
        _lbs.append(_state[0])
        _ubs.append(_state[1])

        children = []

        if len(symbolic.variables) > 2:
            children = symbolic.variables[2:]

        # building a nested And subexpression
        for i, op_r in enumerate(children):

            op_l = names[i]
            and_op = names[i + 1]

            if op_r.is_one or op_r.is_true:

                _coef = {and_op: -4.0, op_l: 2.0}
                _state = (-3.0, 1.0)

            elif op_r.is_zero or op_r.is_false:

                _coef = {and_op: -4.0, op_l: 2.0}
                _state = (-1.0, 3.0)

            else:
                _coef = {and_op: -4.0, op_l: 2.0, op_r.key(): 2.0}
                _state = (-1.0, 3.0)

            _coefs.append(_coef)
            _lbs.append(_state[0])
            _ubs.append(_state[1])

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def linearize_or(symbolic):

        # Following Boolean algebra, an Or (a = b or c) can be translated as: a = b + c - b*c
        # Alternatively, an Or can be written as lb < b + c - a < ub

        # So, for the mid term expression a = b or c
        # We have therefore the equation: -2 <= 2*b + 2*c – 4*a <= 1

        name = symbolic.key()
        names = [f'{name}_{i}' for i, _ in enumerate(symbolic.variables[:-1])]

        or_op = names[0]
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        _coefs = []
        _lbs = []
        _ubs = []

        if op_l.is_one or op_l.is_true:

            _coef = {or_op: -4.0, op_r.key(): 2.0}
            _state = (-4.0, -1.0)

        elif op_r.is_one or op_r.is_true:

            _coef = {or_op: -4.0, op_l.key(): 2.0}
            _state = (-4.0, -1.0)

        elif op_l.is_zero or op_l.is_false:

            _coef = {or_op: -4.0, op_r.key(): 2.0}
            _state = (-2.0, 0.0)

        elif op_r.is_zero or op_r.is_false:

            _coef = {or_op: -4.0, op_l.key(): 2.0}
            _state = (-2.0, 0.0)

        else:

            _coef = {or_op: -4.0, op_l.key(): 2.0, op_r.key(): 2.0}
            _state = (-2.0, 1.0)

        _coefs.append(_coef)
        _lbs.append(_state[0])
        _ubs.append(_state[1])

        children = []

        if len(symbolic.variables) > 2:
            children = symbolic.variables[2:]

        # building a nested Or subexpression
        for i, op_r in enumerate(children):

            op_l = names[i]
            or_op = names[i + 1]

            if op_r.is_one or op_r.is_true:

                _coef = {or_op: -4.0, op_l: 2.0}
                _state = (-4.0, -1.0)

            elif op_r.is_zero or op_r.is_false:

                _coef = {or_op: -4.0, op_l: 2.0}
                _state = (-2.0, 1.0)

            else:

                _coef = {or_op: -4.0, op_l: 2.0, op_r.key(): 2.0}
                _state = (-2.0, 1.0)

            _coefs.append(_coef)
            _lbs.append(_state[0])
            _ubs.append(_state[1])

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def linearize_not(symbolic):

        # Following Boolean algebra, an Not (a = not b) can be translated as: a = 1 - b
        # Alternatively, an Not can be written as a + b = 1

        # So, for the mid term expression a = not b
        # We have therefore the equation: 1 < a + b < 1

        op_l = symbolic.variables[0]

        # Not right operators
        if op_l.is_numeric:

            _coef = {symbolic.key(): 1.0}
            _state = (float(op_l.value), float(op_l.value))

        else:

            # add Not row and set mip bounds to 1;1
            _coef = {symbolic.key(): 1.0, op_l.key(): 1.0}
            _state = (1.0, 1.0)

        return ConstraintContainer(name=None, coefs=[_coef], lbs=[_state[0]], ubs=[_state[1]])

    def linearize_greater(self, symbolic):

        # Following Propositional logic, a predicate (a => r > value) can be translated as: r - value > 0
        # Alternatively, flux predicate a => r>value can be a(value + tolerance - r_UB) + r < value + tolerance
        # & a(r_LB - value - tolerance) + r > r_LB

        # So, for the mid term expression a => r > value
        # We have therefore the equations: a(value + tolerance - r_UB) + r < value + tolerance
        # & a(r_LB - value - tolerance) + r > r_LB

        # In matlab: c * (Vmax - const) + v <= Vmax
        # In matlab: const <= c * (const - Vmin) + v

        # In matlab: compareVal = -compareVal - epsilon
        # In matlab: matrix_values = [vmax - compare_vale, 1] (-inf, vmax)
        # In matlab: matrix_values = [compare_vale - vmin, 1] (compareVal, inf)

        greater_op = symbolic.key()
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        if op_l.is_numeric:
            operand = op_r
            c_val = float(op_l.value)

        else:
            operand = op_l
            c_val = float(op_r.value)

        _lb, _ub = bounds_from_symbol(symbol=operand, model=self.model, default=(MEWPY_LB, MEWPY_UB))

        _lb = float(_lb)
        _ub = float(_ub)
        # add Greater row (a(value + tolerance - r_UB) + r < value + tolerance) and set mip bounds to
        # -inf;comparison_val
        # add Greater row (a(r_LB - value - tolerance) + r > r_LB) and set mip bounds to lb;inf
        _coefs = [
            {greater_op: c_val + MEWPY_TOL - _ub, operand.key(): 1.0},
            {greater_op: _lb - c_val - MEWPY_TOL, operand.key(): 1.0}
        ]

        _lbs = [MEWPY_LB, _lb]
        _ubs = [c_val + MEWPY_TOL, MEWPY_UB]

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    def linearize_less(self, symbolic):

        # Following Propositional logic, a predicate (a => r < value) can be translated as: r - value < 0
        # Alternatively, flux predicate a => r<value can be a(value + tolerance - r_LB) + r > value + tolerance
        # & a(r_UB - value - tolerance) + r < r_UB

        # So, for the mid term expression a => r > value
        # We have therefore the equations: a(value + tolerance - r_LB) + r > value + tolerance
        # & a(r_UB - value - tolerance) + r < r_UB

        # In matlab: c * (const - Vmax) + v <= const
        # In matlab: Vmin <= c * (Vmin - const) + v

        # In matlab: compareVal = -compareVal + epsilon
        # In matlab: matrix_values = [compareVal-vmax, 1] (-inf, compareVal)
        # In matlab: matrix_values = [vmin-compareVal, 1] (vmin, inf)

        less_op = symbolic.key()
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        if op_l.is_numeric:
            operand = op_r
            c_val = float(op_l.value)

        else:
            operand = op_l
            c_val = float(op_r.value)

        _lb, _ub = bounds_from_symbol(symbol=operand, model=self.model, default=(MEWPY_LB, MEWPY_UB))
        _lb = float(_lb)
        _ub = float(_ub)
        # add Less row (a(value + tolerance - r_LB) + r > value + tolerance) and set mip bounds to
        # -inf;-comparison_val
        # add Less row (a(r_UB - value - tolerance) + r < r_UB) and set mip bounds to lb;inf
        _coefs = [
            {less_op: c_val + MEWPY_TOL - _lb, operand.key(): 1.0},
            {less_op: _ub - c_val - MEWPY_TOL, operand.key(): 1.0}
        ]
        _lbs = [c_val + MEWPY_TOL, MEWPY_LB]
        _ubs = [MEWPY_UB, _ub]

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def linearize_equal(symbolic):

        # Following Propositional logic, a predicate (a => r = value) can be translated as: r - value = 0

        less_op = symbolic.key()
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        if op_l.is_numeric:
            operand = op_r
            c_val = float(op_l.value)

        else:
            operand = op_l
            c_val = float(op_r.value)

        _coefs = [{less_op: - c_val, operand.key(): 1.0}]
        _lbs = [0]
        _ubs = [0]

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def linearize_none(symbolic):

        # If the expression is empty, it means that the target or reaction can take any boolean value (0;1)
        # add row and set mip bounds to 0;1

        return ConstraintContainer(name=None, coefs=[{}], lbs=[0.0], ubs=[1.0])

    @staticmethod
    def linearize_false(symbolic):

        return ConstraintContainer(name=None, coefs=[{}], lbs=[0.0], ubs=[0.0])

    @staticmethod
    def linearize_true(symbolic):

        return ConstraintContainer(name=None, coefs=[{}], lbs=[1.0], ubs=[1.0])

    @staticmethod
    def linearize_number(symbolic):

        return ConstraintContainer(name=None, coefs=[{}], lbs=[float(symbolic.value)], ubs=[float(symbolic.value)])

    @staticmethod
    def linearize_symbol(symbolic):

        return ConstraintContainer(name=None, coefs=[{symbolic.key(): -1.0}], lbs=[0.0], ubs=[0.0])

    def get_linearizer(self,
                       symbolic,
                       operators=True,
                       bool_atoms=True,
                       numeric_atoms=True,
                       symbolic_atoms=True,
                       empty_symbolic=True,
                       ):

        if operators:

            if symbolic.is_and:
                return partial(self.linearize_and, symbolic)

            elif symbolic.is_or:
                return partial(self.linearize_or, symbolic)

            elif symbolic.is_not:
                return partial(self.linearize_not, symbolic)

            elif symbolic.is_greater or symbolic.is_greater_equal:
                return partial(self.linearize_greater, symbolic)

            elif symbolic.is_less or symbolic.is_less_equal:
                return partial(self.linearize_less, symbolic)

            elif symbolic.is_equal:
                return partial(self.linearize_equal, symbolic)

        if bool_atoms:
            if symbolic.is_true:

                return partial(self.linearize_true, symbolic)

            elif symbolic.is_false:

                return partial(self.linearize_false, symbolic)

        if numeric_atoms:
            if symbolic.is_numeric:
                return partial(self.linearize_number, symbolic)

        if symbolic_atoms:
            if symbolic.is_symbol:
                return partial(self.linearize_symbol, symbolic)

        if empty_symbolic:
            if symbolic.is_none:
                return partial(self.linearize_none, symbolic)

        return

    @staticmethod
    def _operator_variablize(symbolic):

        name = symbolic.key()
        names = []
        lbs = []
        ubs = []
        var_types = []

        for i, _ in enumerate(symbolic.variables[:-1]):
            names.append(f'{name}_{i}')
            lbs.append(0.0)
            ubs.append(1.0)
            var_types.append(VarType.INTEGER)

        return VariableContainer(name=name, sub_variables=names, lbs=lbs, ubs=ubs, variables_type=var_types)

    def variablize_and(self, symbolic):

        return self._operator_variablize(symbolic=symbolic)

    def variablize_or(self, symbolic):

        return self._operator_variablize(symbolic=symbolic)

    @staticmethod
    def variablize_not(symbolic):

        name = symbolic.key()
        sub_variable_name = f'{name}_0'

        return VariableContainer(name=symbolic.key(),
                                 sub_variables=[sub_variable_name],
                                 lbs=[0.0],
                                 ubs=[1.0],
                                 variables_type=[VarType.INTEGER])

    def variablize_greater(self, symbolic):

        return self._operator_variablize(symbolic=symbolic)

    def variablize_less(self, symbolic):

        return self._operator_variablize(symbolic=symbolic)

    def variablize_equal(self, symbolic):

        return self._operator_variablize(symbolic=symbolic)

    def variablize_symbol(self, symbolic):

        name = symbolic.key()
        lb, ub = bounds_from_symbol(symbol=symbolic, model=self.model)
        lb = float(lb)
        ub = float(ub)

        if (lb, ub) in integer_coefficients:
            v_type = VarType.INTEGER
        else:
            v_type = VarType.CONTINUOUS

        return VariableContainer(name=name, sub_variables=[name], lbs=[lb], ubs=[ub], variables_type=[v_type])

    def get_variabilizer(self, symbolic):

        if symbolic.is_and:
            return partial(self.variablize_and, symbolic)

        elif symbolic.is_or:
            return partial(self.variablize_or, symbolic)

        elif symbolic.is_not:
            return partial(self.variablize_not, symbolic)

        elif symbolic.is_greater or symbolic.is_greater_equal:
            return partial(self.variablize_greater, symbolic)

        elif symbolic.is_less or symbolic.is_less_equal:
            return partial(self.variablize_less, symbolic)

        elif symbolic.is_equal:
            return partial(self.variablize_equal, symbolic)

        elif symbolic.is_symbol:
            return partial(self.variablize_symbol, symbolic)

        return

    def linearize_expression(self, boolean_variable, symbolic):

        # Results of str expression parsing:
        # gene = X and 1 => X
        # gene = X or 1 => 1 (str BooleanTrue)
        # gene = X and 0 => 0 (str BooleanFalse)
        # gene = X or 0 => X
        # gene = 1 => 1 (str one)
        # gene = 0 => 0 (str zero)
        # gene = X > 10 => (X, 1 (str number)), namely two args

        # if expression is atom and defines a variable always On or Off, add a On/Off row
        if symbolic.is_atom:
            return self.linearize_atomic_expression(boolean_variable=boolean_variable, symbolic=symbolic)

        return self.linearize_complex_expression(boolean_variable=boolean_variable, symbolic=symbolic)

    def linearize_atomic_expression(self, boolean_variable, symbolic):

        variables = []
        constraints = []

        if symbolic.is_symbol:
            var = self.variablize_symbol(symbolic=symbolic)
            variables.append(var)

        linearizer = self.get_linearizer(symbolic, operators=False)

        expression_cnt = linearizer()

        expression_cnt.coefs[0][boolean_variable] = 1.0

        constraints.append(expression_cnt)

        return variables, constraints

    def linearize_complex_expression(self, boolean_variable, symbolic):

        # iterating an expression in the reverse order
        # e.g. expr = A and (B or C)
        # list(expression) # [C, B, A, B | C, A & (B | C)]

        variables = []
        constraints = []
        last_variable = None
        # let's add the expressions
        for atom in symbolic:

            # An operator expression will be decomposed into multiple operator expressions, namely into a nested
            # expression. For instance, an A & B & C & D will become And(D, And(C, And(A, B))). Each nested operator
            # expression will be a column in the matrix and then linked in the columns. However, this operator
            # expression will be the same for all maters and have the same column identifier regardless of the length.
            # Thus, all operator columns (variables) will be under the columns linked list engine. When retrieving
            # the operator expression indexes from the cols dictionary in the other methods, the last index
            # of the slice should be used to get the last real column, as the result of this column is the one that
            # really matters. The hashes of the columns for the simulation engine should be the the row name plus
            # str of the operator

            last_variable = atom

            variablizer = self.get_variabilizer(symbolic=atom)

            if variablizer is not None:
                var = variablizer()
                variables.append(var)

            linearizer = self.get_linearizer(atom,
                                             bool_atoms=False,
                                             numeric_atoms=False,
                                             symbolic_atoms=False,
                                             empty_symbolic=False)

            if linearizer is not None:
                constraint = linearizer()
                constraints.append(constraint)

        # identifying the last index to link the outcome of this variable to the boolean variable associated to the
        # expression
        last_variable_name = last_variable.key()
        names = [f'{last_variable_name}_{i}' for i, _ in enumerate(last_variable.variables[:-1])]

        if names:
            last_variable_name = names[-1]
        else:
            last_variable_name = last_variable_name

        # add gene row which means that the gene variable in the mip matrix is associated with the last mid term
        # expression, namely the whole expression
        # set mip bounds to 0;0
        expression_cnt = ConstraintContainer(name=None,
                                             coefs=[{boolean_variable: 1.0, last_variable_name: -1.0}],
                                             lbs=[0.0],
                                             ubs=[0.0])
        constraints.append(expression_cnt)

        return variables, constraints


class GPRLinearizer(LogicLinearizer):

    @abstractmethod
    def build(self):

        # The concrete implementation is defined by each simulation method, e.g. fba, pfba, etc

        pass

    def notification(self, notification: Notification):

        if notification.content_type == 'gprs' and notification.action == 'add':

            return self.add_gprs(notification.content)

        elif notification.content_type == 'gprs' and notification.action == 'remove':

            return self.remove_gprs(notification.content)

        else:
            return super(GPRLinearizer, self).notification(notification)

    def _gpr_constraint(self, reaction: str, symbolic: 'Symbolic'):

        # Relation between reaction boolean value and the reaction constrains
        # For that, the following reactions must be added
        # V - Y*Vmax <= 0
        # V - Y*Vmin => 0
        # where V is the reaction in S matrix
        # where Y is the reaction boolean variable

        # add reaction boolean variable / add column / Add MILP bool variable
        # If already exists, only the index is returned

        boolean_variable = f'bool_{reaction}'

        variables = []
        constraints = []

        boolean_variable_linear_variable = VariableContainer(name=boolean_variable,
                                                             sub_variables=[boolean_variable],
                                                             lbs=[0.0],
                                                             ubs=[1.0],
                                                             variables_type=[VarType.INTEGER])

        variables.append(boolean_variable_linear_variable)

        lb, ub = bounds_from_variable(reaction, model=self.model, default=(MEWPY_LB, MEWPY_UB))

        coefs = [{reaction: 1.0, boolean_variable: -float(ub)},
                 {reaction: 1.0, boolean_variable: -float(lb)}]

        lbs = [MEWPY_LB - float(ub), 0.0]
        ubs = [0.0, MEWPY_UB - float(lb)]

        cnt = ConstraintContainer(name=None,
                                  coefs=coefs,
                                  lbs=lbs,
                                  ubs=ubs)
        constraints.append(cnt)

        expression_variables, expression_cnt = self.linearize_expression(boolean_variable=boolean_variable,
                                                                         symbolic=symbolic)

        variables.extend(expression_variables)
        constraints.extend(expression_cnt)

        constraints = [concat_constraints(constraints=constraints, name=reaction)]

        return variables, constraints

    def add_gprs(self, reactions: List['Reaction']):

        reactions = iterable(reactions)

        to_add = []

        # gprs over reactions
        for rxn in reactions:
            rxn_variables, rxn_cnt = self._gpr_constraint(reaction=rxn.id,
                                                          symbolic=rxn.gpr.symbolic)

            to_add.extend(rxn_variables)
            to_add.extend(rxn_cnt)

        self.add(to_add)

    def remove_gprs(self, reactions: List['Reaction']):

        reactions = iterable(reactions)

        to_remove = []

        for rxn in reactions:
            to_remove.append(self._variables[f'bool_{rxn.id}'])
            to_remove.append(self._constraints[rxn.id])

        self.remove(to_remove)


class InteractionLinearizer(LogicLinearizer):

    @abstractmethod
    def build(self):

        # The concrete implementation is defined by each simulation method, e.g. fba, pfba, etc

        pass

    def notification(self, notification: Notification):

        if notification.content_type == 'interactions' and notification.action == 'add':

            return self.add_interactions(notification.content)

        elif notification.content_type == 'interactions' and notification.action == 'remove':

            return self.remove_interactions(notification.content)

        else:
            return super(InteractionLinearizer, self).notification(notification)

    def _interaction_constraint(self, target, symbolic):

        lb, ub = bounds_from_variable(target, model=self.model, default=(0, 1))
        lb = float(lb)
        ub = float(ub)

        if (lb, ub) in integer_coefficients:
            v_type = VarType.INTEGER
        else:
            v_type = VarType.CONTINUOUS

        variables = []
        constraints = []

        target_linear_variable = VariableContainer(name=target,
                                                   sub_variables=[target],
                                                   lbs=[lb],
                                                   ubs=[ub],
                                                   variables_type=[v_type])

        variables.append(target_linear_variable)

        expression_variables, expression_cnt = self.linearize_expression(boolean_variable=target,
                                                                         symbolic=symbolic)

        variables.extend(expression_variables)
        constraints.extend(expression_cnt)

        constraints = [concat_constraints(constraints=constraints, name=target)]

        return variables, constraints

    def add_interactions(self, interactions):

        interactions = iterable(interactions)

        to_add = []

        # expressions over interactions
        for interaction in interactions:

            symbolic = None

            for coefficient, expression in interaction.regulatory_events.items():

                if coefficient > 0.0:
                    symbolic = expression.symbolic

            if symbolic is None:
                continue

            interaction_variables, interaction_cnt = self._interaction_constraint(target=interaction.target.id,
                                                                                  symbolic=symbolic)

            to_add.extend(interaction_variables)
            to_add.extend(interaction_cnt)

        self.add(to_add)

    def remove_interactions(self, interactions):

        interactions = iterable(interactions)

        to_remove = []

        for interaction in interactions:
            to_remove.append(self._variables[interaction.target.id])
            to_remove.append(self._constraints[interaction.target.id])

        self.remove(to_remove)



from typing import Union

from mewpy.solvers.solver import Solver
from mewpy.mew.lp import Notification, MetabolicLinearizer, InteractionLinearizer, GPRLinearizer
from mewpy.model import Model, MetabolicModel, RegulatoryModel
from mewpy.mew.solution import ModelSolution


# TODO: type hinting and documentation
class SRFBA(MetabolicLinearizer, GPRLinearizer, InteractionLinearizer):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = True,
                 attach: bool = False):
        """
        Steady-state Regulatory Flux Balance Analysis (SRFBA) of a metabolic-regulatory model.
        Implementation of a steady-state version of SRFBA for a integrated metabolic-regulatory model.

        For more details consult Shlomi et al. 2007 at https://dx.doi.org/10.1038%2Fmsb4100141

        :param model: a mewpy Model, MetabolicModel, RegulatoryModel or all. The model is used to retrieve
        variables and constraints to the linear problem

        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used but it will be overwritten
        if build is true.

        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """

        super().__init__(model=model,
                         solver=solver,
                         build=build,
                         attach=attach)

    def build(self):

        if self._variables or self._constraints:
            self.clean()

        if self.model.is_metabolic():
            self.metabolite_reaction_lookup(reactions=self.model.yield_reactions(),
                                            metabolites=self.model.yield_metabolites())

            self.add_gprs(self.model.yield_reactions())

        if self.model.is_regulatory():
            self.add_interactions(self.model.yield_interactions())

        self.update()

        self.set_objective(linear={var.id: value for var, value in self.model.objective.items()},
                           minimize=False)

    def notification(self, notification: Notification):

        if notification.content_type == 'coefficients' and notification.action == 'set':

            # if the notification is a coefficient object, the whole problem must be build again,
            # as lower and upper bounds of variables are also encoded into the constraints

            self.update_milp_constraints(notification.content.id)

            lb, ub = notification.content.bounds

            return super(SRFBA, self).set_bounds(notification.content.id, lb=lb, ub=ub)

        else:
            return super(SRFBA, self).notification(notification)

    def update_milp_constraints(self, regulator_or_target):

        variable = self.model.get(regulator_or_target)

        to_add = []

        if variable.is_regulator():

            for interaction in variable.yield_interactions():
                to_add.append(interaction)

        if variable.is_target():
            to_add.append(variable.interaction)

        self.add_interactions(to_add)

        if variable.is_reaction():
            self.add_gprs([variable])

    def _optimize_with_constraints(self,
                                   objective=None,
                                   minimize=False,
                                   constraints=None,
                                   get_values=True,
                                   shadow_prices=False,
                                   reduced_costs=False,
                                   pool_size=0,
                                   pool_gap=None):

        constraints = constraints.copy()

        old_simulators = self.model.simulators.copy()

        self.model._simulators = [self]

        with self.model:

            for variable, bounds in constraints.items():
                variable_obj = self.model.get(variable)

                # sending notification. Notification processing will update variables and constraints accordingly
                if isinstance(bounds, (tuple, list, set)):
                    variable_obj.coefficient.coefficients = bounds

                else:
                    variable_obj.coefficient.coefficients = (bounds,)

            # updating the last modifications
            self.update()

            # solving for the last modifications
            solver_solution = self._solver.solve(linear=objective,
                                                 quadratic=None,
                                                 minimize=minimize,
                                                 model=None,
                                                 constraints=None,
                                                 get_values=get_values,
                                                 shadow_prices=shadow_prices,
                                                 reduced_costs=reduced_costs,
                                                 pool_size=pool_size,
                                                 pool_gap=pool_gap)

        # context exit. restore. update
        self.update()

        self.model._simulators = old_simulators

        return solver_solution

    def optimize(self,
                 initial_state=None,
                 objective=None,
                 minimize=False,
                 constraints=None,
                 to_solver=False,
                 get_values=True,
                 shadow_prices=False,
                 reduced_costs=False,
                 pool_size=0,
                 pool_gap=None):

        self._set_objective(linear=objective, minimize=minimize)

        if not initial_state:
            initial_state = {}

        if not constraints:
            constraints = {}

        constraints = {**constraints, **initial_state}

        if constraints:
            solver_solution = self._optimize_with_constraints(objective=objective,
                                                              minimize=minimize,
                                                              constraints=constraints,
                                                              get_values=get_values,
                                                              shadow_prices=shadow_prices,
                                                              reduced_costs=reduced_costs,
                                                              pool_size=pool_size,
                                                              pool_gap=pool_gap)

        else:
            solver_solution = self._solver.solve(linear=objective,
                                                 quadratic=None,
                                                 minimize=minimize,
                                                 model=None,
                                                 constraints=None,
                                                 get_values=get_values,
                                                 shadow_prices=shadow_prices,
                                                 reduced_costs=reduced_costs,
                                                 pool_size=pool_size,
                                                 pool_gap=pool_gap)

        if to_solver:
            return solver_solution

        if minimize:
            sense = 'minimize'

        else:
            sense = 'maximize'

        return ModelSolution.from_solver(method=self.method, solution=solver_solution, model=self.model,
                                         objective_direction=sense)



from abc import ABCMeta, abstractmethod


class ModelSolutionInterface(metaclass=ABCMeta):

    """
    Solution interface
    """

    @property
    @abstractmethod
    def objective_direction(self):
        return

    @property
    @abstractmethod
    def method(self):
        return

    @property
    @abstractmethod
    def model(self):
        return

    @property
    @abstractmethod
    def objective(self):
        return

    @property
    @abstractmethod
    def objective_value(self):
        return

    @property
    @abstractmethod
    def status(self):
        return

    @property
    @abstractmethod
    def x(self):
        return

    @property
    @abstractmethod
    def simulator(self):
        return

    @abstractmethod
    def to_summary(self):
        return

    @abstractmethod
    def to_series(self):
        return

    @abstractmethod
    def to_frame(self):
        return

    @abstractmethod
    def __repr__(self):
        return

    @abstractmethod
    def __str__(self):
        return



import numpy as np
import pandas as pd
from itertools import combinations
from ..simulation import get_simulator
from ..simulation.simulation import Simulator
from ..util.parsing import Boolean, GeneEvaluator, build_tree


class ExpressionSet(object):

    def __init__(self, identifiers: list, conditions: list,
                 expression: np.array, p_values: np.array = None):
        """Expression set. The expression values are a numpy array with shape
        (len(identifiers) x len(conditions)).

        Args:
            identifiers (list): Gene or Proteins identifiers
            conditions (list): Time, experiment,... identifiers.
            expression (np.array): expression values.
            p_values (np.array, optional): p-values. Defaults to None.
        """
        n = len(identifiers)
        m = len(conditions)
        if expression.shape != (n, m):
            raise ValueError(
                f"The shape of the expression {expression.shape} does not "
                f"match the expression and conditions sizes ({n},{m})")

        self._identifiers = identifiers
        self._identifier_index = {iden: idx for idx, iden in
                                  enumerate(identifiers)}
        self._conditions = [str(x) for x in conditions]
        self._condition_index = {cond: idx for idx, cond in
                                 enumerate(self._conditions)}
        self._expression = expression
        self._p_values = p_values

    def shape(self):
        """Returns:
            (tuple): the Expression dataset shape
        """
        return self._expression.shape

    def __getitem__(self, item):
        """
        Index the ExpressionSet.
        """
        return self._expression.__getitem__(item)

    def get_condition(self, condition=None, **kwargs):
        """

        Args:
            condition ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """

        if isinstance(condition, int):
            values = self[:, condition]
        elif isinstance(condition, str):
            values = self[:, self._condition_index[condition]]
        else:
            values = self[:, :]

        # format
        form = kwargs.get('format', 'dict')
        if form and condition is not None:
            if form == 'list':
                return values.tolist()
            elif form == 'dict':
                return dict(zip(self._identifiers, values.tolist()))
            else:
                return values
        else:
            return values

    @classmethod
    def from_data_frame(cls, data_frame):
        """Read expression data from a pandas.DataFrame.

        Args:
            data_frame (Dataframe): The expression Dataframe

        Returns:
            ExpressionSet: the expression dataset from the dataframe.
        """
        columns = data_frame.columns.tolist()
        conditions = [c for c in columns if "p-value" not in c]
        p_value_keys = [c for c in columns if "p-value" in c]
        if p_value_keys:
            p_values = data_frame[p_value_keys].values
        else:
            p_values = None

        expression = data_frame[conditions].values
        identifiers = data_frame.index.tolist()
        return ExpressionSet(identifiers, conditions, expression, p_values)

    @classmethod
    def from_csv(cls, file_path: str, **kwargs):
        """Read expression data from a comma separated values (csv) file.

        Args:
            file_path (str): the csv file path.

        Returns:
            ExpressionSet: the expression dataset from the csv file.
        """
        data = pd.read_csv(file_path, **kwargs)
        return cls.from_data_frame(data)

    @property
    def data_frame(self):
        """Build a pandas.DataFrame from the ExpressionProfile.
        Columns headers are conditions and
        line indexes identifiers (genes/proteins)
        """

        if self._p_values is None:
            expression = self._expression
            conditions = self._conditions
        else:
            expression = np.concatenate((self._expression, self.p_values),
                                        axis=1)
            conditions = self._conditions + self.p_value_columns

        return pd.DataFrame(expression,
                            index=self._identifiers,
                            columns=conditions)

    @property
    def p_value_columns(self):
        """ Generate the p-value column names."""
        return [f"{c[0]} {c[1]} p-value"
                for c in combinations(self._conditions, 2)]

    @property
    def p_values(self):
        """Returns the numpy array of p-values.

        Raises:
            ValueError: [description]
        """
        if not self._p_values.all():
            raise ValueError("No p-values defined.")
        else:
            return self._p_values

    @p_values.setter
    def p_values(self, p_values: np.array):
        """Sets p-values

        Args:
            p_values (np.array): [description]

        Raises:
            ValueError: [description]
        """
        if p_values is not None:
            if p_values.shape[1] != len(self.p_value_columns):
                raise ValueError("p-values do not cover all conditions")

        self._p_values = p_values

    @p_values.deleter
    def p_values(self):
        """Delete p_values."""
        self._p_values = None

    def differences(self, p_value=0.005):
        """Calculate the differences based on the MADE method.

        Args:
            p_value (float, optional): [description]. Defaults to 0.005.

        Returns:
            dict: A dictionary of differences
        """

        diff = {}
        for idx, iden in enumerate(self._identifiers):
            diff[iden] = []
            for i in range(1, len(self._conditions)):
                start, end = self._expression[idx, i-1: i+1]
                p_val = self.p_values[idx, i-1]
                if p_val <= p_value:
                    if start < end:
                        diff[iden].append(+1)
                    elif start > end:
                        diff[iden].append(-1)
                    else:
                        diff[iden].append(0)
                else:
                    diff[iden].append(0)
        return diff

    def minmax(self, condition=None):
        """ Return the min and max values for the specified condition.

        Args:
            condition (str): str or int or None, optional (default None)
            The condition to obtain the min and max values for.

        Returns
        -------
        tuple of (min, max)

        """
        values = self.get_condition(condition)
        return np.amin(values), np.amax(values)


def gene_to_reaction_expression(model, gene_exp, and_func=min, or_func=max):
    """Process reaction level from GPRs

    Args:
        model: A model or a MEWpy Simulation
        gene_exp (dict): gene identifiers and expression values
        and_func ([type], optional): Function for AND. Defaults to min.
        or_func ([type], optional): Function for OR. Defaults to max.

    Returns:
        dict: Reaction levels
    """
    if isinstance(model, Simulator):
        sim = model
    else:
        sim = get_simulator(model)

    rxn_exp = {}
    evaluator = GeneEvaluator(gene_exp, and_func, or_func, unexpressed_value=None)
    for rxn_id in sim.reactions:
        gpr = sim.get_gpr(rxn_id)
        if gpr:
            tree = build_tree(gpr, Boolean)
            op_set = tree.get_operands().intersection(set(gene_exp.keys()))
            if len(op_set) == 0:
                lv = None
            else:
                lv = tree.evaluate(evaluator.f_operand, evaluator.f_operator)
            rxn_exp[rxn_id] = lv
    return rxn_exp


class Preprocessing(object):
    """Formulation and implementation of preprocessing decisions.
        (A) Types of gene mapping methods
        (B) Types of thresholding approaches (global and local).
        (C) Formulation of combinations of number of states (Global, Local)
        (D) Decisions about the order in which thresholding and gene mapping
        are performed.
        For Order 1, gene expression is converted to reaction activity followed
        by thresholding of reaction activity;
        For Order 2, thresholding ofgene expression is followed by its
        conversion to reaction activity.

        [1]Anne Richelle,Chintan Joshi,Nathan E. Lewis, Assessing key decisions
        for transcriptomic data integration in biochemical networks, PLOS, 2019
        https://doi.org/10.1371/journal.pcbi.1007185
    """

    def __init__(self, model: Simulator, data: ExpressionSet, **kwargs):
        """[summary]

        Args:
            model ([type]): [description]
            data ([type]): [description]
        """
        self.model = model
        self.data = data
        self._conf = kwargs

    def reactions_expression(self, condition, and_func=None, or_func=None):
        exp = self.data.get_condition(condition, format='dict')
        and_func = self._conf.get(
            'and_func', min) if and_func is None else and_func
        or_func = self._conf.get(
            'or_func', max) if or_func is None else or_func
        rxn_exp = gene_to_reaction_expression(
            self.model, exp, and_func, or_func)
        # Removes None if maybe is none to evaluate GPRs
        res = {k: v for k, v in rxn_exp.items() if v is not None}
        return res

    def percentile(self, condition=None, cutoff=0.25):
        """Processes a percentil threshold and returns the respective
        reaction coefficients, ie, a dictionary of reaction:coeff

        Args:
            condition ([type], optional): [description]. Defaults to None.
            cutoff (float, optional): [description]. Defaults to 0.25.

        Returns:
            dict, float: the coefficients and threshold
        """
        if type(cutoff) is tuple:
            coef = []
            thre = []
            for cut in cutoff:
                rxn_exp = self.reactions_expression(condition)
                threshold = np.percentile(list(rxn_exp.values()), cut)
                coeffs = {r_id: threshold-val for r_id,
                          val in rxn_exp.items() if val < threshold}
                coef.append(coeffs)
                thre.append(threshold)
            coeffs = tuple(coef)
            threshold = tuple(thre)
        else:
            rxn_exp = self.reactions_expression(condition)
            threshold = np.percentile(list(rxn_exp.values()), cutoff)
            coeffs = {r_id: threshold-val for r_id,
                    val in rxn_exp.items() if val < threshold}
        return coeffs, threshold



from enum import Enum

from .csv import RegulatoryCSV, CoExpressionCSV, TargetRegulatorCSV
from .cobra import CobrapyModel, ReframedModel
from .json import JSON
from .sbml import RegulatorySBML, MetabolicSBML


class Engines(Enum):
    """
    List of all engines to read/write files/models

    """

    RegulatoryCSV = RegulatoryCSV
    CoExpressionCSV = CoExpressionCSV
    TargetRegulatorCSV = TargetRegulatorCSV
    RegulatorySBML = RegulatorySBML
    MetabolicSBML = MetabolicSBML
    CobrapyModel = CobrapyModel
    ReframedModel = ReframedModel
    JSON = JSON

    @classmethod
    def has_engine(cls, engine):
        try:
            return cls[engine.__name__]

        except KeyError:

            return False

    @classmethod
    def get(cls, engine, default=None):
        try:
            return cls[engine.__name__]

        except KeyError:

            return default



from abc import abstractmethod, ABC
from collections import OrderedDict
from ..util.parsing import evaluate_expression_tree
from ..util.process import cpu_count
from . import SimulationMethod, SStatus
from joblib import Parallel, delayed
from tqdm import tqdm


class ModelContainer(ABC):
    """Interface for Model container.
       Provides an abstraction from models implementations.
    """

    @property
    def reactions(self):
        """
        :returns: A list of reaction identifiers.

        """
        raise NotImplementedError

    @property
    def genes(self):
        """
        :returns: A list of gene identifiers.

        """
        raise NotImplementedError

    @property
    def proteins(self):
        """
        :returns: A list of proteins identifiers.

        """
        raise NotImplementedError

    @property
    def metabolites(self):
        """
        :returns: A list of metabolite identifiers.

        """
        raise NotImplementedError

    @property
    def compartments(self):
        """
        :returns: A list of compartments identifiers.

        """
        raise NotImplementedError

    @property
    def medium(self):
        raise NotImplementedError

    def get_reaction(self, r_id):
        raise NotImplementedError

    def get_gene(self, g_id):
        raise NotImplementedError

    def get_compartment(self, c_id):
        raise NotImplementedError

    def get_reaction_metabolites(self, r_id):
        rxn = self.get_reaction(r_id)
        return rxn['stoichiometry']

    def get_substrates(self, rxn_id):
        met = self.get_reaction_metabolites(rxn_id)
        return {m_id: coeff for m_id, coeff in met.items() if coeff < 0}

    def get_products(self, rxn_id):
        met = self.get_reaction_metabolites(rxn_id)
        return {m_id: coeff for m_id, coeff in met.items() if coeff > 0}

    def get_exchange_reactions(self):
        return NotImplementedError

    def get_gene_reactions(self):
        raise NotImplementedError

    def get_gpr(self, rxn_id):
        """
        :returns: A string representation of the reaction GPR if exists None otherwise.
        """
        rxn = self.get_reaction(rxn_id)
        return rxn['gpr']

    def summary(self):
        print(f"Metabolites: {len(self.metabolites)}")
        print(f"Reactions: {len(self.reactions)}")
        print(f"Genes: {len(self.genes)}")

    def set_objective(self, reaction):
        raise NotImplementedError


class Simulator(ModelContainer):
    """
    Interface for simulators
    """

    @abstractmethod
    def simulate(self, objective=None, method=SimulationMethod.FBA, maximize=True, constraints=None, reference=None,
                 solver=None, **kwargs):
        """Abstract method to run a phenotype simulation.

        :returns: A SimulationResult.

        """
        raise NotImplementedError

    @abstractmethod
    def FVA(self, obj_frac=0, reactions=None, constraints=None, loopless=False, internal=None, solver=None):
        """ Abstract method to run Flux Variability Analysis (FVA).

        :returns: A dictionary of flux range values.

        """
        raise NotImplementedError

    def simulate_mp(self, objective=None, method=SimulationMethod.FBA, maximize=True, constraints_list=None,
                    reference=None, solver=None, jobs=None, desc="Parallel Simulation", **kwargs):
        """Parallel phenotype simulations.

        :param (dict) objective: The simulations objective. If none, the model objective is considered.
        :param (SimulationMethod) method: The SimulationMethod (FBA, pFBA, lMOMA, etc ...)
        :param (boolean) maximize: The optimization direction
        :param (dict) constraints: A dictionary of constraints to be applied to the model.
        :param (dict) reference: A dictionary of reaction flux values.
        :param (Solver) solver: An instance of the solver.
        :param (int) jobs: The number of parallel jobs.
        :param (str) desc: Description to present in tdqm. 
        """
        constraints_list = [None] if not constraints_list else constraints_list
        jobs = jobs if jobs else cpu_count()
        print(f"Using {jobs} jobs")
        from ..util.utilities import tqdm_joblib
        with tqdm_joblib(tqdm(desc=desc, total=len(constraints_list))) as progress_bar:
            res = Parallel(n_jobs=jobs)(delayed(simulate)(self.model, self.environmental_conditions, objective, method,
                                                          maximize, constraints, reference, solver, **kwargs) for constraints in constraints_list)
        return res

    @abstractmethod
    def get_reaction_bounds(self, r_id):
        raise NotImplementedError

    @abstractmethod
    def metabolite_reaction_lookup(self, force_recalculate=False):
        raise NotImplementedError

    def find(self, pattern=None, sort=False, find_in='r'):
        """A user friendly method to find metabolites, reactions or genes in the model.

        :param pattern: The pattern which can be a regular expression, defaults to None in which case all entries are listed.
        :type pattern: str, optional
        :param sort: if the search results should be sorted, defaults to False
        :type sort: bool, optional
        :param find_in: The search set: 'r' for reactions, 'm' for metabolites, 'g' for genes, defaults to 'r'
        :type find_in: str, optional
        :return: the search results
        :rtype: pandas dataframe
        """
        if find_in == 'm':
            values = self.metabolites
        elif find_in == 'g':
            values = self.genes
        else:
            values = self.reactions
        if pattern:
            import re
            if isinstance(pattern, list):
                patt = '|'.join(pattern)
                re_expr = re.compile(patt)
            else:
                re_expr = re.compile(pattern)
            values = [x for x in values if re_expr.search(x) is not None]
        if sort:
            values.sort()

        import pandas as pd
        if find_in == 'm':
            data = [self.get_metabolite(x) for x in values]
        elif find_in == 'g':
            data = [{'Gene': x} for x in values]
        else:
            data = [self.get_reaction(x) for x in values]
        df = pd.DataFrame(data)
        return df

    def essential_reactions(self, min_growth=0.01):
        """Essential reactions are those when knocked out enable a biomass flux value above a minimal growth defined as
        a percentage of the wild type growth.

        :param float min_growth: Minimal percentage of the wild type growth value. Default 0.01 (1%).
        :returns: A list of essential reactions.
        """
        essential = getattr(self, '_essential_reactions', None)
        if essential is not None:
            return essential
        essential = []
        for rxn in tqdm(self.reactions):
            res = self.simulate(constraints={rxn:  0},  slim=True)
            if not res or res < min_growth:
                essential.append(rxn)
        self._essential_reactions = essential
        return self._essential_reactions

    def evaluate_gprs(self, active_genes):
        """Returns the list of active reactions for a given list of active genes.

        :param list active_genes: List of genes identifiers.
        :returns: A list of active reaction identifiers.

        """
        active_reactions = []
        reactions = self.reactions
        for r_id in reactions:
            gpr = self.get_gpr(r_id)
            if gpr:
                if evaluate_expression_tree(str(gpr), active_genes):
                    active_reactions.append(r_id)
            else:
                active_reactions.append(r_id)
        return active_reactions

    def is_essential_gene(self, gene, min_growth=0.01):
        gr = self.get_gene_reactions()
        active_genes = set(self.genes) - {gene}
        rxns = gr.get(gene, [])
        if not rxns:
            return False
        inactive_reactions = []
        for r_id in rxns:
            gpr = self.get_gpr(r_id)
            if gpr and not evaluate_expression_tree(str(gpr), active_genes):
                inactive_reactions.append(r_id)
        constraints = {rxn: 0 for rxn in inactive_reactions}
        res = self.simulate(constraints=constraints, slim=True)
        return not res or res < min_growth

    def essential_genes(self, min_growth=0.01):
        """Essential genes are those when deleted enable a biomass flux value above a minimal growth defined as
        a percentage of the wild type growth.

        :param float min_growth: Minimal percentage of the wild type growth value. Default 0.01 (1%).
        :returns: A list of essential genes.

        """
        essential = getattr(self, '_essential_genes', None)
        if essential is not None:
            return essential
        essential = []
        for g in tqdm(self.genes):
            if self.is_essential_gene(g):
                essential.append(g)
        self._essential_genes = essential
        return self._essential_genes

    @property
    def reference(self):
        """The reference wild type reaction flux values.

        :returns: A dictionary of wild type reaction flux values.

        """
        ref = getattr(self, '_reference', None)
        if ref is not None:
            return ref
        self._reference = self.simulate(method="pFBA").fluxes
        return self._reference


class SimulationResult(object):
    """Class that represents simulation results and performs operations over them."""

    def __init__(self, model, objective_value, fluxes=None, status=None, envcond=None, model_constraints=None,
                 simul_constraints=None, maximize=True, method=None):
        """
        :param model: A model instance.
        :param objective_value: The phenotype simulation objective value.
        :param dict fluxes: A dictionary of reaction fluxes values.
        :param status: The LP status.
        :param dict envcond: The environmental conditions of the phenotype simulation.
        :param dict model_constraints: Possible persistent additional constraints.
        :param dict simul_constraints: The simulation constraints.
        :param boolean maximize: Optimization direction.

        """
        self.model = model
        self.objective_value = objective_value
        self.status = status
        self.fluxes = fluxes
        self.maximize = maximize
        # Environmental conditions
        self.envcond = envcond if envcond else OrderedDict()
        # Persistent additional constraints not included in the environmental conditions
        self.model_constraints = model_constraints if model_constraints else OrderedDict()
        # Constraints specific to the simulation
        self.simulation_constraints = simul_constraints if simul_constraints else OrderedDict()
        self.method = method

    def get_constraints(self):
        """
        :returns: All constraints applied during the simulation both persistent and simulation specific.
        """
        constraints = OrderedDict()
        constraints.update(self.envcond)
        constraints.update(self.model_constraints)
        constraints.update(self.simulation_constraints)
        return constraints

    def __repr__(self):
        return (f"objective: {self.objective_value}\nStatus: "
                f"{self.status}\nConstraints: {self.get_constraints()}\nMethod:{self.method}")

    def __str__(self):
        return (f"objective: {self.objective_value}\nStatus: "
                f"{self.status}\nConstraints: {self.get_constraints()}\nMethod:{self.method}")

    def find(self, pattern=None, sort=False):
        """Returns a dataframe of reactions and their fluxes matching a pattern or a list of patterns.

        :param pattern: a string or a list of strings. defaults to None
        :param sort: If the dataframe is to be sorted by flux rates. Defaults to False
        :type sort: bool, optional
        :return: returns a dataframe.
        :rtype: pandas.DataFrame
        """
        values = [(key, value) for key, value in self.fluxes.items()]
        if pattern:
            import re
            if isinstance(pattern, list):
                patt = '|'.join(pattern)
                re_expr = re.compile(patt)
            else:
                re_expr = re.compile(pattern)
            values = [x for x in values if re_expr.search(x[0]) is not None]
        if sort:
            values.sort(key=lambda x: x[1])
        import pandas as pd
        df = pd.DataFrame(values, columns=['Reaction ID', 'Flux rate'])
        return df

    @property
    def dataframe(self):
        return self.find()

    def get_net_conversion(self, biomassId=None):
        """Returns a string representation of the net conversion.

        :params str biosmassId: Biomass identifier (optional)
        :returns: A string representation of the net conversion.

        """

        left = ""
        right = ""
        firstLeft, firstRight = True, True
        from . import get_simulator
        sim = get_simulator(self.model)
        ssFluxes = self.fluxes
        for r_id in sim.reactions:
            fluxValue = ssFluxes[r_id]
            sub = list(sim.get_substrates(r_id).keys())
            prod = list(sim.get_products(r_id).keys())
            # if rId is a drain reaction
            if fluxValue != 0.0 and (len(sub) == 0 or len(prod) == 0):
                m = sub + prod
                if fluxValue < 0:
                    if firstLeft:
                        firstLeft = False
                    else:
                        left = left + " + "
                    left = left + str(-1 * fluxValue)
                    left = left + " " + m[0]
                else:
                    if firstRight:
                        firstRight = False
                    else:
                        right = right + " + "
                    right = right + str(fluxValue)
                    right = right + " " + m[0]

        if biomassId and biomassId in ssFluxes.keys():
            biomassFlux = ssFluxes[biomassId]
            if biomassFlux > 0:
                if firstRight:
                    firstRight = False
                else:
                    right = right + " + "
                right = right + str(biomassFlux)
                right = right + " " + biomassId

        return left + " --> " + right

    def get_metabolites_turnover(self):
        """ Calculate metabolite turnover.

        Arguments:
            model: REFRAMED/Cobrapy model or Simulator that generated the solution

        Returns:
            dict: metabolite turnover rates
        """
        from . import get_simulator
        sim = get_simulator(self.model)

        if not self.values:
            return None

        m_r_table = sim.metabolite_reaction_lookup()
        t = {m_id: 0.5*sum([abs(coeff * self.fluxes[r_id]) for r_id, coeff in neighbours.items()])
             for m_id, neighbours in m_r_table.items()}
        return t

    @classmethod
    def fromLinearSolver(cls, solution):
        """Converts a solver Solution object to a SolutionResult object.

        :param solution: solution to be converted
        :type solution: mewpy.solver.solution.Solution
        :raises ValueError: if solution is not an instance of mewpy.solver.solution.Solution
        :return: an instance of SolutionResult
        :rtype: SolutionResult
        """
        from mewpy.solvers import Solution, Status
        smap = {Status.OPTIMAL: SStatus.OPTIMAL,
                Status.UNKNOWN: SStatus.UNKNOWN,
                Status.SUBOPTIMAL: SStatus.SUBOPTIMAL,
                Status.UNBOUNDED: SStatus.UNBOUNDED,
                Status.INFEASIBLE: SStatus.INFEASIBLE,
                Status.INF_OR_UNB: SStatus.INF_OR_UNB
                }
        if not isinstance(solution, Solution):
            raise ValueError('solution should be and instance of mewpy.solvers.solution.Solution')
        return cls(None, solution.fobj, fluxes=solution.values, status=smap[solution.status])


def simulate(model, envcond=None, objective=None, method=SimulationMethod.FBA, maximize=True, constraints=None, reference=None,
             solver=None, **kwargs):
    """Runs an FBA phenotype simulation.

    :param model: cobrapy, reframed, mew constraint-base model
    :param (dict )envcond : Environmental conditions, defaults to None
    :param (dict) objective: the FBA objective , defaults to None
    :param method: The FBA method, defaults to SimulationMethod.FBA
    :param maximize: optimization sense, defaults to True
    :param (dict) constraints: additional constraints , defaults to None
    :param (dict) reference: reference fluxes, defaults to None
    :param solver: a solver instance, defaults to None

    :returns: SimultationResult
    """
    from . import get_simulator
    sim = get_simulator(model, envcond=envcond)
    res = sim.simulate(objective=objective, method=method, maximize=maximize,
                       constraints=constraints, reference=reference,
                       solver=solver, **kwargs)
    return res



import numpy
from inspyred.ec.emo import Pareto

from mewpy.visualization.plot import StreamingPlot
from ..ea import Solution, non_dominated_population


def fitness_statistics(population):
    """Return the basic statistics of the population's fitness values.

    :param population: A population of individuals.

    """

    def minuszero(value):
        return round(value, 6)
            
    stats = {}
    population.sort(reverse=True)
    first = population[0].fitness

    if isinstance(first, Pareto):
        n = len(first.values)
        for i in range(n):
            f = [p.fitness.values[i] for p in population]
            worst_fit = min(f)
            best_fit = max(f)
            med_fit = numpy.median(f)
            avg_fit = numpy.mean(f)
            std_fit = numpy.std(f)
            stats['obj_{}'.format(i)] = {'best': best_fit, 'worst': worst_fit,
                                         'mean': avg_fit, 'median': med_fit, 'std': std_fit}
    else:
        worst_fit = population[-1].fitness
        best_fit = population[0].fitness
        f = [p.fitness for p in population]
        med_fit = numpy.median(f)
        avg_fit = numpy.mean(f)
        std_fit = numpy.std(f)
        stats['obj'] = {'best': minuszero(best_fit), 'worst': minuszero(worst_fit),
                        'mean': minuszero(avg_fit), 'median': minuszero(med_fit), 'std': minuszero(std_fit)}

    return stats


def results_observer(population, num_generations, num_evaluations, args):
    """
    Print the output of the evolutionary computation to a file with the follow fields:
    - number of generation
    - fitness of candidate
    - the solution candidates
    - the solution encoded candidates

    :param population: (list) the population of Individuals.
    :param num_generations: (int) the number of elapsed generations.
    :param num_evaluations: (int) the number of evaluations already performed.
    :param args: (dict) a dictionary of keyword arguments.

    """

    stats = fitness_statistics(population)
    title = "Gen    Eval|"
    values = "{0:>4} {1:>6}|".format(num_generations, num_evaluations)

    for key in stats:
        s = stats[key]
        title = title + "     Worst      Best    Median   Average   Std Dev|"
        values = values + "  {0:.6f}  {1:.6f}  {2:.6f}  {3:.6f}  {4:.6f}|".format(s['worst'],
                                                                                  s['best'],
                                                                                  s['median'],
                                                                                  s['mean'],
                                                                                  s['std'])
    if num_generations == 0:
        print(title)
    print(values)


class VisualizerObserver():

    def __init__(self, reference_front=None, reference_point=None, display_frequency=1, axis_labels=None,
                 non_dominated=False, print_stats=True):
        self.figure = None
        self.display_frequency = display_frequency
        self.reference_point = reference_point
        self.reference_front = reference_front
        self.print_stats = print_stats
        self.axis_labels = axis_labels
        self.non_dominated = non_dominated

    def update(self, population, num_generations, num_evaluations, args):
        generations = num_generations
        evaluations = num_evaluations
        p = []
        for s in population:
            if isinstance(s.fitness, Pareto):
                a = Solution(s.candidate, s.fitness.values)
            else:
                a = Solution(s.candidate, [s.fitness])
            p.append(a)

        nds = non_dominated_population(p)
        ds = None

        if not self.non_dominated:
            ds = list(set(p)-set(nds))

        if self.figure is None:
            self.figure = StreamingPlot(axis_labels=self.axis_labels)
            self.figure.plot(nds, dominated=ds)

        if (generations % self.display_frequency) == 0:
            self.figure.update(nds, dominated=ds)
            self.figure.ax.set_title(
                'Eval: {}'.format(evaluations), fontsize=13)

        if self.print_stats:
            results_observer(population, num_generations,
                             num_evaluations, args)



from .model_solution import ModelSolution
from .multi_solution import MultiSolution, DynamicSolution, KOSolution

# TODO: Can simulation results, EA results and other solutions be moved here?
#  If so, some objects and modules might have to be refactored.
